from __future__ import annotations

from collections.abc import Sequence
from dataclasses import asdict
from typing import TYPE_CHECKING

import ispyb
import ispyb.sqlalchemy
from ispyb.connector.mysqlsp.main import ISPyBMySQLSPConnector as Connector
from ispyb.sp.mxacquisition import MXAcquisition
from ispyb.strictordereddict import StrictOrderedDict
from pydantic import BaseModel

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionGroupInfo,
    DataCollectionInfo,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_utils import (
    get_current_time_string,
    get_session_id_from_visit,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER
from mx_bluesky.common.utils.tracing import TRACER

if TYPE_CHECKING:
    pass

I03_EIGER_DETECTOR = 78
EIGER_FILE_SUFFIX = "h5"


class IspybIds(BaseModel):
    data_collection_ids: tuple[int, ...] = ()
    data_collection_group_id: int | None = None
    grid_ids: tuple[int, ...] = ()


class StoreInIspyb:
    def __init__(self, ispyb_config: str) -> None:
        self.ISPYB_CONFIG_PATH: str = ispyb_config
        self._data_collection_group_id: int | None

    def begin_deposition(
        self,
        data_collection_group_info: DataCollectionGroupInfo,
        scan_data_infos: Sequence[ScanDataInfo],
    ) -> IspybIds:
        ispyb_ids = IspybIds()
        if scan_data_infos[0].data_collection_info:
            ispyb_ids.data_collection_group_id = scan_data_infos[
                0
            ].data_collection_info.parent_id

        return self._begin_or_update_deposition(
            ispyb_ids, data_collection_group_info, scan_data_infos
        )

    def update_deposition(
        self,
        ispyb_ids,
        scan_data_infos: Sequence[ScanDataInfo],
    ) -> IspybIds:
        assert ispyb_ids.data_collection_group_id, (
            "Attempted to store scan data without a collection group"
        )
        assert ispyb_ids.data_collection_ids, (
            "Attempted to store scan data without a collection"
        )
        return self._begin_or_update_deposition(ispyb_ids, None, scan_data_infos)

    def _begin_or_update_deposition(
        self,
        ispyb_ids,
        data_collection_group_info: DataCollectionGroupInfo | None,
        scan_data_infos,
    ) -> IspybIds:
        with ispyb.open(self.ISPYB_CONFIG_PATH) as conn:
            assert conn, "Failed to connect to ISPyB"
            if data_collection_group_info:
                ispyb_ids.data_collection_group_id = (
                    self._store_data_collection_group_table(
                        conn,
                        data_collection_group_info,
                        ispyb_ids.data_collection_group_id,
                    )
                )
            else:
                assert ispyb_ids.data_collection_group_id, (
                    "Attempt to update data collection without a data collection group ID"
                )

            grid_ids = list(ispyb_ids.grid_ids)
            data_collection_ids_out = list(ispyb_ids.data_collection_ids)
            for scan_data_info in scan_data_infos:
                data_collection_id = scan_data_info.data_collection_id
                if (
                    scan_data_info.data_collection_info
                    and not scan_data_info.data_collection_info.parent_id
                ):
                    scan_data_info.data_collection_info.parent_id = (
                        ispyb_ids.data_collection_group_id
                    )

                new_data_collection_id, grid_id = self._store_single_scan_data(
                    conn, scan_data_info, data_collection_id
                )
                if not data_collection_id:
                    data_collection_ids_out.append(new_data_collection_id)
                if grid_id:
                    grid_ids.append(grid_id)
            ispyb_ids = IspybIds(
                data_collection_ids=tuple(data_collection_ids_out),
                grid_ids=tuple(grid_ids),
                data_collection_group_id=ispyb_ids.data_collection_group_id,
            )
        return ispyb_ids

    def end_deposition(self, ispyb_ids: IspybIds, success: str, reason: str):
        assert ispyb_ids.data_collection_ids, (
            "Can't end ISPyB deposition, data_collection IDs are missing"
        )
        assert ispyb_ids.data_collection_group_id is not None, (
            "Cannot end ISPyB deposition without data collection group ID"
        )

        for id_ in ispyb_ids.data_collection_ids:
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                f"End ispyb deposition with status '{success}' and reason '{reason}'."
            )
            if success == "fail" or success == "abort":
                run_status = "DataCollection Unsuccessful"
            else:
                run_status = "DataCollection Successful"
            current_time = get_current_time_string()
            self._update_scan_with_end_time_and_status(
                current_time,
                run_status,
                reason,
                id_,
                ispyb_ids.data_collection_group_id,
            )

    def append_to_comment(
        self, data_collection_id: int, comment: str, delimiter: str = " "
    ) -> None:
        with ispyb.open(self.ISPYB_CONFIG_PATH) as conn:
            assert conn is not None, "Failed to connect to ISPyB!"
            mx_acquisition: MXAcquisition = conn.mx_acquisition
            mx_acquisition.update_data_collection_append_comments(
                data_collection_id, comment, delimiter
            )

    def update_data_collection_group_table(
        self,
        dcg_info: DataCollectionGroupInfo,
        data_collection_group_id: int | None = None,
    ) -> None:
        with ispyb.open(self.ISPYB_CONFIG_PATH) as conn:
            assert conn is not None, "Failed to connect to ISPyB!"
            self._store_data_collection_group_table(
                conn,
                dcg_info,
                data_collection_group_id,
            )

    def _update_scan_with_end_time_and_status(
        self,
        end_time: str,
        run_status: str,
        reason: str,
        data_collection_id: int,
        data_collection_group_id: int,
    ) -> None:
        if reason is not None and reason != "":
            self.append_to_comment(data_collection_id, f"{run_status} reason: {reason}")

        with ispyb.open(self.ISPYB_CONFIG_PATH) as conn:
            assert conn is not None, "Failed to connect to ISPyB!"

            mx_acquisition: MXAcquisition = conn.mx_acquisition

            params = mx_acquisition.get_data_collection_params()
            params["id"] = data_collection_id
            params["parentid"] = data_collection_group_id
            params["endtime"] = end_time
            params["run_status"] = run_status

            mx_acquisition.upsert_data_collection(list(params.values()))

    def _store_position_table(
        self, conn: Connector, dc_pos_info, data_collection_id
    ) -> int:
        mx_acquisition: MXAcquisition = conn.mx_acquisition

        params = mx_acquisition.get_dc_position_params()
        params["id"] = data_collection_id
        params |= asdict(dc_pos_info)

        return mx_acquisition.update_dc_position(list(params.values()))

    def _store_data_collection_group_table(
        self,
        conn: Connector,
        dcg_info: DataCollectionGroupInfo,
        data_collection_group_id: int | None = None,
    ) -> int:
        mx_acquisition: MXAcquisition = conn.mx_acquisition

        params = mx_acquisition.get_data_collection_group_params()
        if data_collection_group_id:
            params["id"] = data_collection_group_id
        params["parent_id"] = get_session_id_from_visit(conn, dcg_info.visit_string)
        params |= {k: v for k, v in asdict(dcg_info).items() if k != "visit_string"}

        return self._upsert_data_collection_group(conn, params)

    def _store_data_collection_table(
        self, conn, data_collection_id, data_collection_info
    ):
        if data_collection_id and data_collection_info.comments:
            self.append_to_comment(
                data_collection_id, data_collection_info.comments, " "
            )
            data_collection_info.comments = None

        params = self._fill_common_data_collection_params(
            conn, data_collection_id, data_collection_info
        )

        return self._upsert_data_collection(conn, params)

    def _store_single_scan_data(
        self, conn, scan_data_info, data_collection_id=None
    ) -> tuple[int, int | None]:
        data_collection_id = self._store_data_collection_table(
            conn, data_collection_id, scan_data_info.data_collection_info
        )

        if scan_data_info.data_collection_position_info:
            self._store_position_table(
                conn,
                scan_data_info.data_collection_position_info,
                data_collection_id,
            )

        grid_id = None
        if scan_data_info.data_collection_grid_info:
            grid_id = self._store_grid_info_table(
                conn,
                data_collection_id,
                scan_data_info.data_collection_grid_info,
            )
        return data_collection_id, grid_id

    def _store_grid_info_table(
        self,
        conn: Connector,
        ispyb_data_collection_id: int,
        dc_grid_info: DataCollectionGridInfo,
    ) -> int:
        mx_acquisition: MXAcquisition = conn.mx_acquisition
        params = mx_acquisition.get_dc_grid_params()
        params |= dc_grid_info.as_dict()
        params["parentid"] = ispyb_data_collection_id
        return mx_acquisition.upsert_dc_grid(list(params.values()))

    def _fill_common_data_collection_params(
        self, conn, data_collection_id, data_collection_info: DataCollectionInfo
    ) -> StrictOrderedDict:
        mx_acquisition: MXAcquisition = conn.mx_acquisition
        params = mx_acquisition.get_data_collection_params()

        if data_collection_id:
            params["id"] = data_collection_id
        if data_collection_info.visit_string:
            # This is only needed for populating the DataCollectionGroup
            params["visit_id"] = get_session_id_from_visit(
                conn, data_collection_info.visit_string
            )
        params |= {
            k: v for k, v in asdict(data_collection_info).items() if k != "visit_string"
        }

        return params

    @staticmethod
    @TRACER.start_as_current_span("_upsert_data_collection_group")
    def _upsert_data_collection_group(
        conn: Connector, params: StrictOrderedDict
    ) -> int:
        return conn.mx_acquisition.upsert_data_collection_group(list(params.values()))

    @staticmethod
    @TRACER.start_as_current_span("_upsert_data_collection")
    def _upsert_data_collection(conn: Connector, params: StrictOrderedDict) -> int:
        return conn.mx_acquisition.upsert_data_collection(list(params.values()))



import json
from pathlib import Path
from unittest.mock import ANY, MagicMock, call, mock_open, patch

import bluesky.plan_stubs as bps
import pytest
from dodal.devices.i24.beamstop import Beamstop
from dodal.devices.i24.dual_backlight import DualBacklight
from dodal.devices.i24.i24_detector_motion import DetectorMotion
from dodal.devices.i24.pmac import PMAC
from ophyd_async.testing import get_mock_put

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import Fiducials
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1 import (
    cs_maker,
    cs_reset,
    fiducial,
    initialise_stages,
    laser_control,
    moveto,
    moveto_preset,
    pumpprobe_calc,
    read_parameters,
    scrape_mtr_directions,
    scrape_mtr_fiducials,
    set_pmac_strings_for_cs,
    upload_chip_map_to_geobrick,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger

chipmap_str = """01status    P3011       1
02status    P3021       0
03status    P3031       0
04status    P3041       0"""

mtr_dir_str = """#Some words
mtr1_dir=1
mtr2_dir=-1
mtr3_dir=-1"""

fiducial_1_str = """MTR RBV Corr
MTR1 0 1
MTR2 1 -1
MTR3 0 -1"""

cs_json = '{"scalex":1, "scaley":2, "scalez":3, "skew":-0.5, "Sx_dir":1, "Sy_dir":-1, "Sz_dir":0}'


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.get_detector_type"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.get_chip_format"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1._read_visit_directory_from_file"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.SSX_LOGGER"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.bps.rd")
def test_read_parameters(
    fake_rd,
    fake_log,
    mock_read_visit,
    fake_caget,
    fake_chip,
    fake_det,
    detector_stage,
    RE,
):
    def fake_generator(value):
        yield from bps.null()
        return value

    mock_attenuator = MagicMock()
    fake_det.side_effect = [fake_generator(Eiger())]
    fake_rd.side_effect = [fake_generator(0.3)]
    with patch(
        "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.FixedTargetParameters",
    ):
        RE(read_parameters(detector_stage, mock_attenuator))

    assert fake_caget.call_count == 12
    assert fake_log.info.call_count == 3


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.sys")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.get_detector_type"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
async def test_initialise(
    fake_caput: MagicMock,
    fake_det: MagicMock,
    fake_sys: MagicMock,
    pmac: PMAC,
    RE,
):
    RE(initialise_stages(pmac))

    assert await pmac.x.velocity.get_value() == 20
    assert await pmac.y.acceleration_time.get_value() == 0.01
    assert await pmac.z.high_limit_travel.get_value() == 5.1
    assert await pmac.z.low_limit_travel.get_value() == -4.1

    mock_pmac_str = get_mock_put(pmac.pmac_string)
    mock_pmac_str.assert_has_calls(
        [
            call("m508=100 m509=150", wait=True),
            call("m608=100 m609=150", wait=True),
            call("m708=100 m709=150", wait=True),
            call("m808=100 m809=150", wait=True),
        ]
    )


@pytest.mark.parametrize(
    "fake_chip_map",
    [[10], [1, 2, 15, 16], list(range(33, 65))],  # 1 block, 1 corner, half chip
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.bps.sleep"
)
def test_upload_chip_map_to_geobrick(
    fake_sleep: MagicMock, fake_chip_map: list[int], pmac: PMAC, RE
):
    tot_blocks = 64
    RE(upload_chip_map_to_geobrick(pmac, fake_chip_map))

    mock_pmac_str = get_mock_put(pmac.pmac_string)
    assert mock_pmac_str.call_count == tot_blocks

    pvar_zero_calls = [
        call(f"P3{i:02d}1=0", wait=True) for i in range(1, 65) if i not in fake_chip_map
    ]

    pvar_one_calls = [call(f"P3{i:02d}1=1", wait=True) for i in fake_chip_map]

    assert len(pvar_zero_calls) == tot_blocks - len(fake_chip_map)
    assert len(pvar_one_calls) == len(fake_chip_map)

    mock_pmac_str.assert_has_calls(pvar_one_calls, any_order=True)
    mock_pmac_str.assert_has_calls(pvar_zero_calls, any_order=True)


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
async def test_moveto_oxford_origin(fake_caget: MagicMock, pmac: PMAC, RE):
    fake_caget.return_value = 0
    RE(moveto(Fiducials.origin, pmac))
    assert fake_caget.call_count == 1
    assert await pmac.x.user_setpoint.get_value() == 0.0
    assert await pmac.y.user_setpoint.get_value() == 0.0


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
async def test_moveto_oxford_inner_f1(fake_caget: MagicMock, pmac: PMAC, RE):
    fake_caget.return_value = 1
    RE(moveto(Fiducials.fid1, pmac))
    assert fake_caget.call_count == 1
    assert await pmac.x.user_setpoint.get_value() == 24.60
    assert await pmac.y.user_setpoint.get_value() == 0.0


async def test_moveto_chip_aspecific(pmac: PMAC, RE):
    RE(moveto("zero", pmac))
    assert await pmac.pmac_string.get_value() == "!x0y0z0"


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
async def test_moveto_preset(
    fake_caput: MagicMock,
    pmac: PMAC,
    beamstop: Beamstop,
    backlight: DualBacklight,
    detector_stage: DetectorMotion,
    RE,
):
    RE(moveto_preset("zero", pmac, beamstop, backlight, detector_stage))
    assert await pmac.pmac_string.get_value() == "!x0y0z0"

    RE(moveto_preset("load_position", pmac, beamstop, backlight, detector_stage))
    assert await beamstop.pos_select.get_value() == "Robot"
    assert await backlight.backlight_position.pos_level.get_value() == "Out"
    assert await detector_stage.z.user_setpoint.get_value() == 1300


@pytest.mark.parametrize(
    "pos_request, expected_num_caput, expected_pmac_move, other_devices",
    [
        ("collect_position", 1, [0.0, 0.0, 0.0], True),
        ("microdrop_position", 0, [6.0, -7.8, 0.0], False),
    ],
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
async def test_moveto_preset_with_pmac_move(
    fake_caput: MagicMock,
    pos_request: str,
    expected_num_caput: int,
    expected_pmac_move: list,
    other_devices: bool,
    pmac: PMAC,
    beamstop: Beamstop,
    backlight: DualBacklight,
    detector_stage: DetectorMotion,
    RE,
):
    RE(moveto_preset(pos_request, pmac, beamstop, backlight, detector_stage))
    assert fake_caput.call_count == expected_num_caput

    assert await pmac.x.user_setpoint.get_value() == expected_pmac_move[0]
    assert await pmac.y.user_setpoint.get_value() == expected_pmac_move[1]
    assert await pmac.z.user_setpoint.get_value() == expected_pmac_move[2]

    if other_devices:
        assert await beamstop.pos_select.get_value() == "Data Collection"
        assert await backlight.backlight_position.pos_level.get_value() == "In"


@pytest.mark.parametrize(
    "laser_setting, expected_pmac_string",
    [
        ("laser1on", " M712=1 M711=1"),
        ("laser1off", " M712=0 M711=1"),
        ("laser2on", " M812=1 M811=1"),
        ("laser2off", " M812=0 M811=1"),
    ],
)
async def test_laser_control_on_and_off(
    laser_setting: str, expected_pmac_string: str, pmac: PMAC, RE
):
    RE(laser_control(laser_setting, pmac))

    assert await pmac.pmac_string.get_value() == expected_pmac_string


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.bps.sleep"
)
def test_laser_control_burn_setting(
    fake_sleep: MagicMock, fake_caget: MagicMock, pmac: PMAC, RE
):
    fake_caget.return_value = 0.1
    RE(laser_control("laser1burn", pmac))

    fake_sleep.assert_called_once_with(0.1)
    mock_pmac_str = get_mock_put(pmac.pmac_string)
    mock_pmac_str.assert_has_calls(
        [
            call(" M712=1 M711=1", wait=True),
            call(" M712=0 M711=1", wait=True),
        ]
    )


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data=mtr_dir_str),
)
def test_scrape_mtr_directions():
    res = scrape_mtr_directions()
    assert len(res) == 3
    assert res == (1.0, -1.0, -1.0)


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.bps.rd")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.PARAM_FILE_PATH_FT"
)
def test_fiducial_writes_correct_values_to_file(
    fake_param_path, patch_read, patch_mtr, pmac, RE
):
    fake_param_path.return_value = Path("/tmp/params")
    mtr_values = (1.0, -1.0, -1.0)
    patch_mtr.return_value = mtr_values

    pos = (1.02, 4.5, 0.0)

    def fake_generator(value):
        yield from bps.null()
        return value

    patch_read.side_effect = [
        fake_generator(pos[0]),
        fake_generator(pos[1]),
        fake_generator(pos[2]),
    ]

    expected_write_calls = [
        call(fake_param_path / "fiducial_1.txt", "w"),
        call().write("MTR\tRBV\tCorr\n"),
        call().write(f"MTR1\t{pos[0]:1.4f}\t{mtr_values[0]:f}\n"),
        call().write(f"MTR2\t{pos[1]:1.4f}\t{mtr_values[1]:f}\n"),
        call().write(f"MTR3\t{pos[2]:1.4f}\t{mtr_values[2]:f}"),
    ]
    with patch(
        "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
        mock_open(),
    ) as mock_file:
        RE(fiducial(1, pmac))

        mock_file.assert_has_calls(expected_write_calls, any_order=True)
        fake_param_path.mkdir.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data=fiducial_1_str),
)
def test_scrape_mtr_fiducials():
    res = scrape_mtr_fiducials(1)
    assert len(res) == 3
    assert res == (0.0, 1.0, 0.0)


def test_cs_pmac_str_set(pmac: PMAC, RE):
    RE(
        set_pmac_strings_for_cs(
            pmac,
            {
                "cs1": "#1->-10000X+0Y+0Z",
                "cs2": "#2->+0X+10000Y+0Z",
                "cs3": "#3->0X+0Y+10000Z",
            },
        )
    )
    mock_pmac_str = get_mock_put(pmac.pmac_string)
    mock_pmac_str.assert_has_calls(
        [
            call("&2", wait=True),
            call("#1->-10000X+0Y+0Z", wait=True),
            call("#2->+0X+10000Y+0Z", wait=True),
            call("#3->0X+0Y+10000Z", wait=True),
        ]
    )


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.set_pmac_strings_for_cs"
)
def test_cs_reset(mock_set_pmac_str: MagicMock, pmac: PMAC, RE):
    RE(cs_reset(pmac))
    mock_set_pmac_str.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data='{"a":11, "b":12,}'),
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_fiducials"
)
def test_cs_maker_raises_error_for_invalid_json(
    fake_fid: MagicMock,
    fake_dir: MagicMock,
    fake_caget: MagicMock,
    pmac: PMAC,
    RE,
):
    fake_dir.return_value = (1, 1, 1)
    fake_fid.return_value = (0, 0, 0)
    with pytest.raises(json.JSONDecodeError):
        RE(cs_maker(pmac))


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data='{"scalex":11, "skew":12}'),
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_fiducials"
)
def test_cs_maker_raises_error_for_missing_key_in_json(
    fake_fid: MagicMock,
    fake_dir: MagicMock,
    fake_caget: MagicMock,
    pmac: PMAC,
    RE,
):
    fake_dir.return_value = (1, 1, 1)
    fake_fid.return_value = (0, 0, 0)
    with pytest.raises(KeyError):
        RE(cs_maker(pmac))


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data=cs_json),
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_fiducials"
)
def test_cs_maker_raises_error_for_wrong_direction_in_json(
    fake_fid: MagicMock,
    fake_dir: MagicMock,
    fake_caget: MagicMock,
    pmac: PMAC,
    RE,
):
    fake_dir.return_value = (1, 1, 1)
    fake_fid.return_value = (0, 0, 0)
    with pytest.raises(ValueError):
        RE(cs_maker(pmac))


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
def test_pumpprobe_calc(fake_caget: MagicMock, fake_caput: MagicMock, RE):
    fake_caget.side_effect = [0.01, 0.005]
    RE(pumpprobe_calc())
    assert fake_caget.call_count == 2
    assert fake_caput.call_count == 5
    fake_caput.assert_has_calls(
        [
            call(ANY, 0.62),
            call(ANY, 1.24),
            call(ANY, 1.86),
            call(ANY, 3.1),
            call(ANY, 6.2),
        ]
    )



from __future__ import annotations

import functools
import json
import os
import threading
from collections.abc import Callable
from dataclasses import dataclass
from sys import argv
from time import sleep
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from blueapi.core import BlueskyContext
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.zebra.zebra import Zebra
from flask.testing import FlaskClient

from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.__main__ import (
    Actions,
    BlueskyRunner,
    Status,
    create_app,
    setup_context,
)
from mx_bluesky.hyperion.experiment_plans.experiment_registry import PLAN_REGISTRY
from mx_bluesky.hyperion.parameters.cli import parse_cli_args
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

from ...conftest import raw_params_from_file
from ..conftest import mock_beamline_module_filepaths

FGS_ENDPOINT = "/flyscan_xray_centre/"
START_ENDPOINT = FGS_ENDPOINT + Actions.START.value
STOP_ENDPOINT = Actions.STOP.value
STATUS_ENDPOINT = Actions.STATUS.value
SHUTDOWN_ENDPOINT = Actions.SHUTDOWN.value
TEST_BAD_PARAM_ENDPOINT = "/fgs_real_params/" + Actions.START.value
TEST_PARAMS = json.dumps(
    raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_parameters.json"
    )
)

SECS_PER_RUNENGINE_LOOP = 0.1
RUNENGINE_TAKES_TIME_TIMEOUT = 15

"""
Every test in this file which uses the test_env fixture should either:
    - set RE_takes_time to false
    or
    - set an error on the mock run engine
In order to avoid threads which get left alive forever after test completion
"""


autospec_patch = functools.partial(patch, autospec=True, spec_set=True)


class MockRunEngine:
    def __init__(self, test_name):
        self.RE_takes_time = True
        self.aborting_takes_time = False
        self.error: Exception | None = None
        self.test_name = test_name

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        time = 0.0
        while self.RE_takes_time:
            sleep(SECS_PER_RUNENGINE_LOOP)
            time += SECS_PER_RUNENGINE_LOOP
            if self.error:
                raise self.error
            if time > RUNENGINE_TAKES_TIME_TIMEOUT:
                raise TimeoutError(
                    f'Mock RunEngine thread for test "{self.test_name}" spun too long'
                    "without an error. Most likely you should initialise with "
                    "RE_takes_time=false, or set RE.error from another thread."
                )
        if self.error:
            raise self.error

    def abort(self):
        while self.aborting_takes_time:
            sleep(SECS_PER_RUNENGINE_LOOP)
            if self.error:
                raise self.error
        self.RE_takes_time = False

    def subscribe(self, *args):
        pass

    def unsubscribe(self, *args):
        pass


@dataclass
class ClientAndRunEngine:
    client: FlaskClient
    mock_run_engine: MockRunEngine


def mock_dict_values(d: dict):
    return {k: MagicMock() if k == "setup" or k == "run" else v for k, v in d.items()}


TEST_EXPTS = {
    "test_experiment": {
        "setup": MagicMock(),
        "param_type": MagicMock(),
    },
    "fgs_real_params": {
        "setup": MagicMock(),
        "param_type": HyperionSpecifiedThreeDGridScan,
    },
}


@pytest.fixture
def test_env(request: pytest.FixtureRequest):
    mock_run_engine = MockRunEngine(test_name=repr(request))
    mock_context = BlueskyContext()
    real_plans_and_test_exps = dict(
        {k: mock_dict_values(v) for k, v in PLAN_REGISTRY.items()},  # type: ignore
        **TEST_EXPTS,  # type: ignore
    )
    mock_context.plan_functions = {  # type: ignore
        k: MagicMock() for k in real_plans_and_test_exps.keys()
    }

    with (
        patch.dict(
            "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
            real_plans_and_test_exps,
        ),
        patch(
            "mx_bluesky.hyperion.__main__.setup_context",
            MagicMock(return_value=mock_context),
        ),
    ):
        app, runner = create_app({"TESTING": True}, mock_run_engine, True)  # type: ignore

    runner_thread = threading.Thread(target=runner.wait_on_queue)
    runner_thread.start()
    with (
        app.test_client() as client,
        patch.dict(
            "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
            real_plans_and_test_exps,
        ),
    ):
        yield ClientAndRunEngine(client, mock_run_engine)

    runner.shutdown()
    runner_thread.join(timeout=3)
    del mock_run_engine


def wait_for_run_engine_status(
    client: FlaskClient,
    status_check: Callable[[str], bool] = lambda status: status != Status.BUSY.value,
    attempts=10,
):
    while attempts != 0:
        response = client.get(STATUS_ENDPOINT)
        response_json = json.loads(response.data)
        LOGGER.debug(
            f"Checking client status - response: {response_json}, attempts left={attempts}"
        )
        if status_check(response_json["status"]):
            return response_json
        else:
            attempts -= 1
            sleep(0.2)
    raise AssertionError("Run engine still busy")


def check_status_in_response(response_object, expected_result: Status):
    response_json = json.loads(response_object.data)
    assert response_json["status"] == expected_result.value, (
        f"{response_json['status']} != {expected_result.value}: {response_json.get('message')}"
    )


def test_start_gives_success(test_env: ClientAndRunEngine):
    response = test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    check_status_in_response(response, Status.SUCCESS)


def test_getting_status_return_idle(test_env: ClientAndRunEngine):
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    test_env.client.put(STOP_ENDPOINT)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.IDLE)


def test_getting_status_after_start_sent_returns_busy(
    test_env: ClientAndRunEngine,
):
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.BUSY)


def test_putting_bad_plan_fails(test_env: ClientAndRunEngine):
    response = test_env.client.put("/bad_plan/start", data=TEST_PARAMS).json
    assert isinstance(response, dict)
    assert response.get("status") == Status.FAILED.value
    assert (
        response.get("message")
        == "PlanNotFound(\"Experiment plan 'bad_plan' not found in registry.\")"
    )
    test_env.mock_run_engine.abort()


def test_plan_with_no_params_fails(test_env: ClientAndRunEngine):
    response = test_env.client.put(
        "/test_experiment_no_internal_param_type/start", data=TEST_PARAMS
    ).json
    assert isinstance(response, dict)
    assert response.get("status") == Status.FAILED.value
    assert isinstance(message := response.get("message"), str)
    assert "'test_experiment_no_internal_param_type' not found in registry." in message
    test_env.mock_run_engine.abort()


def test_sending_start_twice_fails(test_env: ClientAndRunEngine):
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    response = test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    check_status_in_response(response, Status.FAILED)


def test_given_started_when_stopped_then_success_and_idle_status(
    test_env: ClientAndRunEngine,
):
    test_env.mock_run_engine.aborting_takes_time = True
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    response = test_env.client.put(STOP_ENDPOINT)
    check_status_in_response(response, Status.ABORTING)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.ABORTING)
    test_env.mock_run_engine.aborting_takes_time = False
    wait_for_run_engine_status(
        test_env.client, lambda status: status != Status.ABORTING
    )
    check_status_in_response(response, Status.ABORTING)


def test_given_started_when_stopped_and_started_again_then_runs(
    test_env: ClientAndRunEngine,
):
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    test_env.client.put(STOP_ENDPOINT)
    test_env.mock_run_engine.RE_takes_time = True
    response = test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    check_status_in_response(response, Status.SUCCESS)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.BUSY)
    test_env.mock_run_engine.RE_takes_time = False


def test_when_started_n_returnstatus_interrupted_bc_RE_aborted_thn_error_reptd(
    test_env: ClientAndRunEngine,
):
    test_env.mock_run_engine.aborting_takes_time = True
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    test_env.client.put(STOP_ENDPOINT)
    test_env.mock_run_engine.error = Exception("D'Oh")
    response_json = wait_for_run_engine_status(
        test_env.client, lambda status: status != Status.ABORTING.value
    )
    assert response_json["status"] == Status.FAILED.value
    assert response_json["message"] == 'Exception("D\'Oh")'
    assert response_json["exception_type"] == "Exception"


@pytest.mark.parametrize(
    "endpoint, test_file",
    [
        [
            START_ENDPOINT,
            "tests/test_data/parameter_json_files/good_test_parameters.json",
        ],
        [
            "/grid_detect_then_xray_centre/start",
            "tests/test_data/parameter_json_files/good_test_grid_with_edge_detect_parameters.json",
        ],
        [
            "/rotation_scan/start",
            "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json",
        ],
        [
            "/pin_tip_centre_then_xray_centre/start",
            "tests/test_data/parameter_json_files/good_test_pin_centre_then_xray_centre_parameters.json",
        ],
        [
            "/multi_rotation_scan/start",
            "tests/test_data/parameter_json_files/good_test_multi_rotation_scan_parameters.json",
        ],
        [
            "/load_centre_collect_full/start",
            "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json",
        ],
    ],
)
def test_start_with_json_file_gives_success(
    test_env: ClientAndRunEngine, endpoint: str, test_file: str
):
    test_env.mock_run_engine.RE_takes_time = False

    with open(test_file) as test_params_file:
        test_params = test_params_file.read()
    response = test_env.client.put(endpoint, data=test_params)
    check_status_in_response(response, Status.SUCCESS)


def test_start_with_json_file_with_extras_gives_error(test_env: ClientAndRunEngine):
    test_env.mock_run_engine.RE_takes_time = False

    with open(
        "tests/test_data/parameter_json_files/good_test_parameters.json"
    ) as test_params_file:
        test_params = test_params_file.read()

    params = json.loads(test_params)
    params["extra_param"] = "test"
    test_params = json.dumps(params)
    response = test_env.client.put(START_ENDPOINT, data=test_params)
    check_status_in_response(response, Status.FAILED)


test_argument_combinations = [
    (
        [
            "--dev",
        ],
        (True, False, False, False),
    ),
    ([], (False, False, False, False)),
    (
        [
            "--dev",
            "--skip-startup-connection",
            "--verbose-event-logging",
        ],
        (True, True, True),
    ),
]


@pytest.mark.parametrize(["arg_list", "parsed_arg_values"], test_argument_combinations)
def test_cli_args_parse(arg_list, parsed_arg_values):
    argv[1:] = arg_list
    test_args = parse_cli_args()
    assert test_args.dev_mode == parsed_arg_values[0]
    assert test_args.verbose_event_logging == parsed_arg_values[1]
    assert test_args.skip_startup_connection == parsed_arg_values[2]


@pytest.mark.skip(
    "Wait for connection doesn't play nice with ophyd-async. See https://github.com/DiamondLightSource/hyperion/issues/1159"
)
def test_when_blueskyrunner_initiated_then_plans_are_setup_and_devices_connected():
    zebra = MagicMock(spec=Zebra)
    attenuator = MagicMock(spec=BinaryFilterAttenuator)

    context = BlueskyContext()
    context.register_device(zebra, "zebra")
    context.register_device(attenuator, "attenuator")

    @dataclass
    class FakeComposite:
        attenuator: BinaryFilterAttenuator
        zebra: Zebra

    # A fake setup for a plan that uses two devices: attenuator and zebra.
    def fake_create_devices(context) -> FakeComposite:
        print("CREATING DEVICES")
        return device_composite_from_context(context, FakeComposite)

    with patch.dict(
        "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
        {
            "flyscan_xray_centre": {
                "setup": fake_create_devices,
                "param_type": MagicMock(),
            },
        },
        clear=True,
    ):
        print(PLAN_REGISTRY)

        BlueskyRunner(
            RE=MagicMock(),
            context=context,
            skip_startup_connection=False,
        )

    zebra.wait_for_connection.assert_called()
    attenuator.wait_for_connection.assert_called()


@patch(
    "mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan.create_devices",
    autospec=True,
)
def test_when_blueskyrunner_initiated_and_skip_flag_is_set_then_setup_called_upon_start(
    mock_setup, test_fgs_params: HyperionSpecifiedThreeDGridScan
):
    mock_setup = MagicMock()
    with patch.dict(
        "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
        {
            "flyscan_xray_centre": {
                "setup": mock_setup,
                "param_type": MagicMock(),
            },
        },
        clear=True,
    ):
        runner = BlueskyRunner(MagicMock(), MagicMock(), skip_startup_connection=True)
        mock_setup.assert_not_called()
        runner.start(lambda: None, test_fgs_params, "flyscan_xray_centre")
        mock_setup.assert_called_once()
        runner.shutdown()


def test_when_blueskyrunner_initiated_and_skip_flag_is_not_set_then_all_plans_setup():
    mock_setup = MagicMock()
    with patch.dict(
        "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
        {
            "flyscan_xray_centre": {
                "setup": mock_setup,
                "param_type": MagicMock(),
            },
            "rotation_scan": {
                "setup": mock_setup,
                "param_type": MagicMock(),
            },
            "other_plan": {
                "setup": mock_setup,
                "param_type": MagicMock(),
            },
            "yet_another_plan": {
                "setup": mock_setup,
                "param_type": MagicMock(),
            },
        },
        clear=True,
    ):
        BlueskyRunner(MagicMock(), MagicMock(), skip_startup_connection=False)
        assert mock_setup.call_count == 4


def test_log_on_invalid_json_params(test_env: ClientAndRunEngine):
    test_env.mock_run_engine.RE_takes_time = False
    response = test_env.client.put(TEST_BAD_PARAM_ENDPOINT, data='{"bad":1}').json
    assert isinstance(response, dict)
    assert response.get("status") == Status.FAILED.value
    assert (message := response.get("message")) is not None
    assert message.startswith(
        "ValueError('Supplied parameters don\\'t match the plan for this endpoint"
    )
    assert response.get("exception_type") == "ValueError"


@pytest.mark.skip(
    reason="See https://github.com/DiamondLightSource/hyperion/issues/777"
)
def test_warn_exception_during_plan_causes_warning_in_log(
    caplog: pytest.LogCaptureFixture, test_env: ClientAndRunEngine
):
    test_env.client.put(START_ENDPOINT, data=TEST_PARAMS)
    test_env.mock_run_engine.error = WarningException("D'Oh")
    response_json = wait_for_run_engine_status(test_env.client)
    assert response_json["status"] == Status.FAILED.value
    assert response_json["message"] == 'WarningException("D\'Oh")'
    assert response_json["exception_type"] == "WarningException"
    assert caplog.records[-1].levelname == "WARNING"


@patch(
    "dodal.devices.undulator_dcm.get_beamline_parameters",
    return_value={"DCM_Perp_Offset_FIXED": 111},
)
def test_when_context_created_then_contains_expected_number_of_plans(
    get_beamline_parameters,
):
    from dodal.beamlines import i03

    mock_beamline_module_filepaths("i03", i03)

    with patch.dict(
        os.environ,
        {"BEAMLINE": "i03"},
    ):
        context = setup_context(wait_for_connection=False)

        plan_names = context.plans.keys()

        assert "rotation_scan" in plan_names
        assert "grid_detect_then_xray_centre" in plan_names
        assert "pin_tip_centre_then_xray_centre" in plan_names



from __future__ import annotations

import bluesky.plan_stubs as bps
from dodal.devices.aperturescatterguard import (
    ApertureScatterguard,
    ApertureValue,
)
from dodal.devices.backlight import Backlight, BacklightPosition
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.smargon import Smargon

from mx_bluesky.common.utils.log import LOGGER

LOWER_DETECTOR_SHUTTER_AFTER_SCAN = True


def setup_sample_environment(
    aperture_scatterguard: ApertureScatterguard,
    aperture_position_gda_name: str | None,
    backlight: Backlight,
    group="setup_senv",
):
    """Move the aperture into required position, move out the backlight."""
    aperture_value = (
        None
        if not aperture_position_gda_name
        else ApertureValue(aperture_position_gda_name)
    )
    yield from move_aperture_if_required(
        aperture_scatterguard, aperture_value, group=group
    )
    yield from bps.abs_set(backlight, BacklightPosition.OUT, group=group)


def move_aperture_if_required(
    aperture_scatterguard: ApertureScatterguard,
    aperture_value: ApertureValue | None,
    group="move_aperture",
):
    if not aperture_value:
        previous_aperture_position = yield from bps.rd(aperture_scatterguard)
        assert isinstance(previous_aperture_position, ApertureValue)
        LOGGER.info(
            f"Using previously set aperture position {previous_aperture_position}"
        )

    else:
        LOGGER.info(f"Setting aperture position to {aperture_value}")
        yield from bps.abs_set(
            aperture_scatterguard,
            aperture_value,
            group=group,
        )


def cleanup_sample_environment(
    detector_motion: DetectorMotion,
    group="cleanup_senv",
):
    """Put the detector shutter back down"""

    yield from bps.abs_set(
        detector_motion.shutter,
        int(not LOWER_DETECTOR_SHUTTER_AFTER_SCAN),
        group=group,
    )


def move_x_y_z(
    smargon: Smargon,
    x_mm: float | None = None,
    y_mm: float | None = None,
    z_mm: float | None = None,
    wait=False,
    group="move_x_y_z",
):
    """Move the x, y, and z axes of the given smargon to the specified position. All
    axes are optional."""

    LOGGER.info(f"Moving smargon to x, y, z: {(x_mm, y_mm, z_mm)}")
    if x_mm is not None:
        yield from bps.abs_set(smargon.x, x_mm, group=group)
    if y_mm is not None:
        yield from bps.abs_set(smargon.y, y_mm, group=group)
    if z_mm is not None:
        yield from bps.abs_set(smargon.z, z_mm, group=group)
    if wait:
        yield from bps.wait(group)


def move_phi_chi_omega(
    smargon: Smargon,
    phi: float | None = None,
    chi: float | None = None,
    omega: float | None = None,
    wait=False,
    group="move_phi_chi_omega",
):
    """Move the x, y, and z axes of the given smargon to the specified position. All
    axes are optional."""

    LOGGER.info(f"Moving smargon to phi, chi, omega: {(phi, chi, omega)}")
    if phi is not None:
        yield from bps.abs_set(smargon.phi, phi, group=group)
    if chi is not None:
        yield from bps.abs_set(smargon.chi, chi, group=group)
    if omega is not None:
        yield from bps.abs_set(smargon.omega, omega, group=group)
    if wait:
        yield from bps.wait(group)



import logging
from logging.handlers import TimedRotatingFileHandler
from os import environ
from pathlib import Path

from dodal.log import (
    ERROR_LOG_BUFFER_LINES,
    CircularMemoryHandler,
    DodalLogHandlers,
    integrate_bluesky_and_ophyd_logging,
    set_up_all_logging_handlers,
)
from dodal.log import LOGGER as dodal_logger

LOGGER = logging.getLogger("MX-Bluesky")
LOGGER.setLevel("DEBUG")
LOGGER.parent = dodal_logger

ISPYB_ZOCALO_CALLBACK_LOGGER = logging.getLogger("ISPyB and Zocalo callbacks")
ISPYB_ZOCALO_CALLBACK_LOGGER.setLevel(logging.DEBUG)

NEXUS_LOGGER = logging.getLogger("NeXus callbacks")
NEXUS_LOGGER.setLevel(logging.DEBUG)

ALL_LOGGERS = [LOGGER, ISPYB_ZOCALO_CALLBACK_LOGGER, NEXUS_LOGGER]

__logger_handlers: DodalLogHandlers | None = None


class ExperimentMetadataTagFilter(logging.Filter):
    """When an instance of this custom filter is added to a logging handler, dc_group_id
    and run_id will be tagged in that handlers' log messages."""

    dc_group_id: str | None = None
    run_uid: str | None = None

    def filter(self, record):
        if self.dc_group_id:
            record.dc_group_id = self.dc_group_id
        if self.run_uid:
            record.run_uid = self.run_uid
        return True


tag_filter = ExperimentMetadataTagFilter()


def set_dcgid_tag(dcgid):
    """Set the datacollection group id as a tag on all subsequent log messages.
    Setting to None will remove the tag."""
    tag_filter.dc_group_id = dcgid


def set_uid_tag(uid):
    """Set the unique id as a tag on all subsequent log messages.
    Setting to None will remove the tag."""
    tag_filter.run_uid = uid


def do_default_logging_setup(
    file_name: str,
    graylog_port: int,
    dev_mode: bool = False,
    integrate_all_logs: bool = True,
):
    """Configures dodal logger so that separate debug and info log files are created,
    info logs are sent to Graylog, info logs are streamed to sys.sterr, and logs from ophyd
    and bluesky and ophyd-async are optionally included."""

    handlers = set_up_all_logging_handlers(
        dodal_logger,
        _get_logging_dir(),
        file_name,
        dev_mode,
        ERROR_LOG_BUFFER_LINES,
        graylog_port,
    )

    if integrate_all_logs:
        integrate_bluesky_and_ophyd_logging(dodal_logger)

    handlers["graylog_handler"].addFilter(tag_filter)

    global __logger_handlers
    __logger_handlers = handlers


def _get_debug_handler() -> CircularMemoryHandler:
    assert __logger_handlers is not None, (
        "You can only use this after running the default logging setup"
    )
    return __logger_handlers["debug_memory_handler"]


def flush_debug_handler() -> str:
    """Writes the contents of the circular debug log buffer to disk and returns the written filename"""
    handler = _get_debug_handler()
    assert isinstance(handler.target, TimedRotatingFileHandler), (
        "Circular memory handler doesn't have an appropriate fileHandler target"
    )
    handler.flush()
    return handler.target.baseFilename


def _get_logging_dir() -> Path:
    """Get the path to write the mx_bluesky log files to.

    Log location can be specified in the LOG_DIR environment variable, otherwise MX bluesky logs are written to 'dls_sw/ixx/logs/bluesky'.
    This directory will be created if it is not found

    Logs are written to ./tmp/logs/bluesky if BEAMLINE environment variable is not found

    Returns:
        logging_path (Path): Path to the log file for the file handler to write to.
    """

    logging_str = environ.get("LOG_DIR")
    if logging_str:
        logging_path = Path(logging_str)
    else:
        beamline = environ.get("BEAMLINE")
        logging_path = (
            Path(f"/dls_sw/{beamline}/logs/bluesky/")
            if beamline
            else Path("/tmp/logs/bluesky")
        )
    Path.mkdir(logging_path, exist_ok=True, parents=True)
    return logging_path



import datetime
import json
import math
import os
import subprocess
from functools import lru_cache

import bluesky.plan_stubs as bps
import requests
from dodal.beamlines import i24
from dodal.devices.i24.beam_center import DetectorBeamCenter
from dodal.devices.i24.dcm import DCM
from dodal.devices.i24.focus_mirrors import FocusMirrorsMode
from dodal.devices.i24.pilatus_metadata import PilatusMetadata

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import PumpProbeSetting
from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER
from mx_bluesky.beamlines.i24.serial.parameters import (
    BeamSettings,
    DetectorName,
    ExtruderParameters,
    FixedTargetParameters,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import Detector, Eiger, Pilatus

# Collection start/end script to kick off analysis
COLLECTION_START_SCRIPT = "/dls_sw/i24/scripts/RunAtStartOfCollect-i24-ssx.sh"
COLLECTION_END_SCRIPT = "/dls_sw/i24/scripts/RunAtEndOfCollect-i24-ssx.sh"

DEFAULT_ISPYB_SERVER = "https://ssx-dcserver.diamond.ac.uk"

CREDENTIALS_LOCATION = "/scratch/ssx_dcserver.key"


@lru_cache(maxsize=1)
def get_auth_header() -> dict:
    """Read the credentials file and build the Authorisation header"""
    if not os.path.isfile(CREDENTIALS_LOCATION):
        SSX_LOGGER.warning(
            "Could not read %s; attempting to proceed without credentials",
            CREDENTIALS_LOCATION,
        )
        return {}
    with open(CREDENTIALS_LOCATION) as f:
        token = f.read().strip()
    return {"Authorization": "Bearer " + token}


def read_beam_info_from_hardware(
    dcm: DCM,
    mirrors: FocusMirrorsMode,
    beam_center: DetectorBeamCenter,
    detector_name: DetectorName,
):
    """ Read the beam information from hardware.

    Args:
        dcm (DCM): The decm device.
        mirrors (FocusMirrorMode): The device describing the focus mirror mode settings.
        beam_center (DetectorBeamCenter): A device to set and read the beam center on \
            the detector.
        detector_name (DetectorName): The detector currently in use.

    Returns:
        BeamSettings parameter model.
    """
    wavelength = yield from bps.rd(dcm.wavelength_in_a)
    beamsize_x = yield from bps.rd(mirrors.beam_size_x)
    beamsize_y = yield from bps.rd(mirrors.beam_size_y)
    pixel_size = (
        Eiger().pixel_size_mm if detector_name == "eiger" else Pilatus().pixel_size_mm
    )
    beam_center_x = yield from bps.rd(beam_center.beam_x)
    beam_center_y = yield from bps.rd(beam_center.beam_y)
    return BeamSettings(
        wavelength_in_a=wavelength,
        beam_size_in_um=(beamsize_x, beamsize_y),
        beam_center_in_mm=(
            beam_center_x * pixel_size[0],
            beam_center_y * pixel_size[1],
        ),
    )


class DCID:
    """ Interfaces with ISPyB to allow ssx DCID/synchweb interaction.

    Args:
        server (str, optional): The URL for the bridge server, if not the default.
        emit_errors (bool, optional): If False, errors while interacting with the DCID \
            server will not be propagated to the caller. This decides if you want to \
            stop collection if you can't get a DCID. Defaults to True.
        timeout (float, optional): Length of time in s to wait for the DB server before \
            giving up. Defaults to 10 s.
        expt_parameters (ExtruderParameters | FixedTargetParameters): Collection \
            parameters input by user.


    Attributes:
        error:
            If an error has occured. This will be set, even if emit_errors = True
    """

    def __init__(
        self,
        *,
        server: str | None = None,
        emit_errors: bool = True,
        timeout: float = 10,
        expt_params: ExtruderParameters | FixedTargetParameters,
    ):
        self.parameters = expt_params
        self.detector: Detector
        # Handle case of string literal
        match expt_params.detector_name:
            case "eiger":
                self.detector = Eiger()
            case "pilatus":
                self.detector = Pilatus()

        self.server = server or DEFAULT_ISPYB_SERVER
        self.emit_errors = emit_errors
        self.error = False
        self.timeout = timeout
        self.dcid = None

    def generate_dcid(
        self,
        beam_settings: BeamSettings,
        image_dir: str,
        file_template: str,
        num_images: int,
        shots_per_position: int = 1,
        start_time: datetime.datetime | None = None,
        pump_probe: bool = False,
    ):
        """Generate an ispyb DCID.

        Args:
            beam_settings (BeamSettings): Information about the beam read from hardware.
            image_dir (str): The location the images will be written to.
            num_images (int): Total number of images to be collected.
            shots_per_position (int, optional): Number of exposures per position in a \
                chip. Defaults to 1, which works for extruder.
            start_time(datetime, optional): Collection start time. Defaults to None.
            pump_probe (bool, optional): If True, a pump probe collection is running. \
                Defaults to False.
        """
        try:
            if not start_time:
                start_time = datetime.datetime.now().astimezone()
            elif not start_time.timetz:
                start_time = start_time.astimezone()

            resolution = get_resolution(
                self.detector,
                self.parameters.detector_distance_mm,
                beam_settings.wavelength_in_a,
            )
            beamsize_x, beamsize_y = beam_settings.beam_size_in_um
            transmission = self.parameters.transmission * 100
            xbeam, ybeam = beam_settings.beam_center_in_mm

            if isinstance(self.detector, Pilatus):
                startImageNumber = 0
            elif isinstance(self.detector, Eiger):
                startImageNumber = 1
            else:
                raise ValueError("Unknown detector:", self.detector)

            events = [
                {
                    "name": "Xray probe",
                    "offset": 0,
                    "duration": self.parameters.exposure_time_s,
                    "period": self.parameters.exposure_time_s,
                    "repetition": shots_per_position,
                    "eventType": "XrayDetection",
                }
            ]
            if pump_probe:
                match self.parameters:
                    case FixedTargetParameters():
                        # pump then probe - pump_delay corresponds to time *before* first image
                        pump_delay = (
                            -self.parameters.laser_delay_s
                            if self.parameters.pump_repeat
                            is not PumpProbeSetting.Short2
                            else self.parameters.laser_delay_s
                        )
                    case ExtruderParameters():
                        pump_delay = self.parameters.laser_delay_s
                events.append(
                    {
                        "name": "Laser probe",
                        "offset": pump_delay,
                        "duration": self.parameters.laser_dwell_s,
                        "repetition": 1,
                        "eventType": "LaserExcitation",
                    },
                )

            data = {
                "detectorDistance": self.parameters.detector_distance_mm,
                "detectorId": self.detector.id,
                "exposureTime": self.parameters.exposure_time_s,
                "fileTemplate": file_template,
                "imageDirectory": image_dir,
                "numberOfImages": num_images,
                "resolution": resolution,
                "startImageNumber": startImageNumber,
                "startTime": start_time.isoformat(),
                "transmission": transmission,
                "visit": self.parameters.visit.name,
                "wavelength": beam_settings.wavelength_in_a,
                "group": {
                    "experimentType": self.parameters.ispyb_experiment_type.value
                },
                "xBeam": xbeam,
                "yBeam": ybeam,
                "ssx": {
                    "eventChain": {
                        "events": events,
                    }
                },
            }
            if beamsize_x and beamsize_y:
                data["beamSizeAtSampleX"] = beamsize_x / 1000
                data["beamSizeAtSampleY"] = beamsize_y / 1000

            # Log what we are doing here
            try:
                SSX_LOGGER.info(
                    "BRIDGE: POST /dc --data %s",
                    repr(json.dumps(data)),
                )
            except Exception:
                SSX_LOGGER.info(
                    "Caught exception converting data to JSON. Data:\n%s\nVERBOSE:\n%s",
                    str({k: type(v) for k, v in data.items()}),
                )
                raise

            resp = requests.post(
                f"{self.server}/dc",
                json=data,
                timeout=self.timeout,
                headers=get_auth_header(),
            )
            resp.raise_for_status()
            self.dcid = resp.json()["dataCollectionId"]
            SSX_LOGGER.info("Generated DCID %s", self.dcid)
        except requests.HTTPError as e:
            self.error = True
            SSX_LOGGER.error(
                "DCID generation Failed; Reason from server: %s", e.response.text
            )
            if self.emit_errors:
                raise
            SSX_LOGGER.exception("Error generating DCID: %s", e)
        except Exception as e:
            self.error = True
            if self.emit_errors:
                raise
            SSX_LOGGER.exception("Error generating DCID: %s", e)

    def __int__(self):
        return self.dcid

    def notify_start(self):
        """Send notifications that the collection is now starting"""
        if self.dcid is None:
            return None
        try:
            command = [COLLECTION_START_SCRIPT, str(self.dcid)]
            SSX_LOGGER.info("Running %s", " ".join(command))
            subprocess.Popen(command)
        except Exception as e:
            self.error = True
            if self.emit_errors:
                raise
            SSX_LOGGER.warning("Error starting start of collect script: %s", e)

    def notify_end(self):
        """Send notifications that the collection has now ended"""
        if self.dcid is None:
            return
        try:
            command = [COLLECTION_END_SCRIPT, str(self.dcid)]
            SSX_LOGGER.info("Running %s", " ".join(command))
            subprocess.Popen(command)
        except Exception as e:
            self.error = True
            if self.emit_errors:
                raise
            SSX_LOGGER.warning("Error running end of collect notification: %s", e)

    def collection_complete(
        self, end_time: str | datetime.datetime | None = None, aborted: bool = False
    ) -> None:
        """
        Mark an ispyb DCID as completed.

        Args:
            dcid: The Collection ID to mark as finished
            end_time: The predetermined end time
            aborted: Was this collection aborted?
        """
        try:
            # end_time might be a string from time.ctime
            if isinstance(end_time, str):
                end_time = datetime.datetime.strptime(end_time, "%a %b %d %H:%M:%S %Y")
                SSX_LOGGER.debug("Parsed end time: %s", end_time)

            if not end_time:
                end_time = datetime.datetime.now().astimezone()
            if not end_time.tzinfo:
                end_time = end_time.astimezone()

            status = (
                "DataCollection Cancelled" if aborted else "DataCollection Successful"
            )
            data = {
                "endTime": end_time.isoformat(),
                "runStatus": status,
            }
            if self.dcid is None:
                # Print what we would have sent. This means that if something is failing,
                # we still have the data to upload in the log files.
                SSX_LOGGER.info(
                    'BRIDGE: No DCID but Would PATCH "/dc/XXXX" --data=%s',
                    repr(json.dumps(data)),
                )
                return

            SSX_LOGGER.info(
                'BRIDGE: PATCH "/dc/%s" --data=%s', self.dcid, repr(json.dumps(data))
            )
            response = requests.patch(
                f"{self.server}/dc/{self.dcid}",
                json=data,
                timeout=self.timeout,
                headers=get_auth_header(),
            )
            response.raise_for_status()
            SSX_LOGGER.info("Successfully updated end time for DCID %d", self.dcid)
        except Exception as e:
            resp_obj = getattr(e, "response", None)
            try:
                if resp_obj is not None:
                    resp_str = resp_obj.text
                # resp_str = repr(getattr(e, "Iresponse", "<no attribute>"))
                else:
                    resp_str = "Resp object is None"
            except Exception:
                resp_str = f"<failed to determine {resp_obj!r}>"

            self.error = True
            if self.emit_errors:
                raise
            SSX_LOGGER.warning("Error completing DCID: %s (%s)", e, resp_str)


def get_pilatus_filename_template_from_device():
    """
    Get the template file path by querying the detector PVs, mirror the construction \
    that the PPU does.

    Returns:
        A template string, with the image numbers replaced with '#'
    """
    pilatus_metadata: PilatusMetadata = i24.pilatus_metadata()

    filename_template = yield from bps.rd(pilatus_metadata.filename_template)
    return filename_template


def get_resolution(detector: Detector, distance: float, wavelength: float) -> float:
    """ Calculate the inscribed resolution for detector.

    This assumes perfectly centered beam as I don't know where to extract the beam \
    position parameters yet.

    Args:
        detector (Detector): Detector instance, Eiger() or Pilatus().
        distance (float): Distance to detector, in mm.
        wavelength (float): Beam wavelength, in Å.

    Returns:
        Maximum resolution, in Å.
    """
    width = detector.image_size_mm[0]
    return round(wavelength / (2 * math.sin(math.atan(width / (2 * distance)) / 2)), 2)



import io
import json
import pickle
from datetime import timedelta
from typing import TypedDict

import numpy as np
import zmq
from dodal.beamlines.i04 import MURKO_REDIS_DB, REDIS_HOST, REDIS_PASSWORD
from numpy.typing import NDArray
from PIL import Image
from redis import StrictRedis

from mx_bluesky.common.utils.log import LOGGER

MURKO_ADDRESS = "tcp://i04-murko-prod.diamond.ac.uk:8008"

MurkoResult = dict
FullMurkoResults = dict[str, list[MurkoResult]]


class MurkoRequest(TypedDict):
    """See https://github.com/MartinSavko/murko#usage for more information."""

    to_predict: NDArray
    model_img_size: tuple[int, int]
    save: bool
    min_size: int
    description: list

    # The identifier for each image
    prefix: list[str]


def get_image_size(image: NDArray) -> tuple[int, int]:
    """Returns the width and height of a numpy image"""
    return image.shape[1], image.shape[0]


def send_to_murko_and_get_results(request: MurkoRequest) -> FullMurkoResults:
    LOGGER.info(f"Sending {request['prefix']} to murko")
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect(MURKO_ADDRESS)
    socket.send(pickle.dumps(request))
    raw_results = socket.recv()
    assert isinstance(raw_results, bytes)
    results = pickle.loads(raw_results)
    LOGGER.info(f"Got {len(results['descriptions'])} results")
    return results


def _correlate_results_to_uuids(
    request: MurkoRequest, murko_results: FullMurkoResults
) -> list[tuple[str, MurkoResult]]:
    """We send a batch of images to murko, with each having a 'prefix' of the uuid that
    we're using to keep track of the image. Murko sends back an ordered list of these,
    which we match to the supplied prefix here."""
    return list(zip(request["prefix"], murko_results["descriptions"], strict=False))


class BatchMurkoForwarder:
    def __init__(self, redis_client: StrictRedis, batch_size: int):
        """
        Holds image data streamed from redis and forwards it to murko when:
            * A set number have been received
            * The shape of the images changes
            * When `flush` is called

        Once data has been forwarded this will then wait on the results and put them
        back in redis.

        Args:
            redis_client: The client to send murko results back to redis.
            batch_size: How many results to accumulate until they are flushed to redis.
        """
        self.redis_client = redis_client
        self.batch_size = batch_size
        self._uuids_and_images: dict[str, NDArray] = {}
        self._last_image_size: tuple[int, int] | None = None
        self._last_sample_id = ""

    def _handle_batch_of_images(self, sample_id, images, uuids):
        request_arguments: MurkoRequest = {
            "model_img_size": (256, 320),
            "to_predict": np.array(images),
            "save": False,
            "min_size": 64,
            "description": [
                "foreground",
                "crystal",
                "loop_inside",
                "loop",
                ["crystal", "loop"],
                ["crystal", "loop", "stem"],
            ],
            "prefix": uuids,
        }
        results = send_to_murko_and_get_results(request_arguments)
        results_with_uuids = _correlate_results_to_uuids(request_arguments, results)
        self._send_murko_results_to_redis(sample_id, results_with_uuids)

    def _send_murko_results_to_redis(
        self, sample_id: str, results: list[tuple[str, MurkoResult]]
    ):
        """Stores the results into a redis hash (for longer term storage) and publishes
        them as well so that downstream clients can get notified."""
        for uuid, result in results:
            redis_key = f"murko:{sample_id}:results"
            self.redis_client.hset(redis_key, uuid, str(pickle.dumps(result)))
            self.redis_client.expire(redis_key, timedelta(days=7))
        self.redis_client.publish("murko-results", pickle.dumps(results))

    def add(self, sample_id: str, uuid: str, image: NDArray):
        """Add an image to the batch to send to murko."""
        image_size = get_image_size(image)
        self._last_sample_id = sample_id
        if self._last_image_size and self._last_image_size != image_size:
            self.flush()
        self._uuids_and_images[uuid] = image
        self._last_image_size = image_size
        if len(self._uuids_and_images.keys()) >= self.batch_size:
            self.flush()

    def flush(self):
        """Flush the batch to murko."""
        if self._uuids_and_images:
            self._handle_batch_of_images(
                self._last_sample_id,
                list(self._uuids_and_images.values()),
                list(self._uuids_and_images.keys()),
            )
        self._uuids_and_images = {}
        self._last_image_size = None


class RedisListener:
    TIMEOUT_S = 2

    def __init__(
        self,
        redis_host=REDIS_HOST,
        redis_password=REDIS_PASSWORD,
        db=MURKO_REDIS_DB,
        redis_channel="murko",
    ):
        self.redis_client = StrictRedis(
            host=redis_host,
            password=redis_password,
            db=db,
        )
        self.pubsub = self.redis_client.pubsub()
        self.channel = redis_channel
        self.forwarder = BatchMurkoForwarder(self.redis_client, 10)

    def _get_and_handle_message(self):
        message = self.pubsub.get_message(timeout=self.TIMEOUT_S)
        if message and message["type"] == "message":
            data = json.loads(message["data"])
            LOGGER.info(f"Received from redis: {data}")
            uuid = data["uuid"]
            sample_id = data["sample_id"]

            # Images are put in redis as raw jpeg bytes, murko needs numpy arrays
            image_key = f"murko:{sample_id}:raw"
            raw_image = self.redis_client.hget(image_key, uuid)

            if not isinstance(raw_image, bytes):
                LOGGER.warning(
                    f"Image at {image_key}:{uuid} is {raw_image}, expected bytes. Ignoring the data"
                )
                return

            image = Image.open(io.BytesIO(raw_image))
            image = np.asarray(image)

            self.forwarder.add(sample_id, uuid, image)

        elif not message:
            self.forwarder.flush()

    def listen_for_image_data_forever(self):
        self.pubsub.subscribe(self.channel)

        while True:
            self._get_and_handle_message()


def main():
    client = RedisListener()
    client.listen_for_image_data_forever()


if __name__ == "__main__":
    main()



import pytest
from dodal.devices.zebra.zebra import (
    TrigSource,
    Zebra,
)

from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_zebra_plans import (
    arm_zebra,
    disarm_zebra,
    get_zebra_settings_for_extruder,
    open_fast_shutter_at_each_position_plan,
    reset_output_panel,
    reset_pc_gate_and_pulse,
    reset_zebra_when_collection_done_plan,
    set_shutter_mode,
    setup_pc_sources,
    setup_zebra_for_extruder_with_pump_probe_plan,
    setup_zebra_for_fastchip_plan,
    setup_zebra_for_quickshot_plan,
    zebra_return_to_normal_plan,
)


async def test_arm_and_disarm_zebra(zebra: Zebra, RE):
    zebra.pc.arm.TIMEOUT = 1

    RE(arm_zebra(zebra))
    assert await zebra.pc.is_armed()

    RE(disarm_zebra(zebra))
    assert await zebra.pc.is_armed() is False


async def test_set_shutter_mode(zebra: Zebra, RE):
    RE(set_shutter_mode(zebra, "manual"))
    assert await zebra.inputs.soft_in_1.get_value() == "No"


async def test_setup_pc_sources(zebra: Zebra, RE):
    RE(setup_pc_sources(zebra, TrigSource.TIME, TrigSource.POSITION))

    assert await zebra.pc.gate_source.get_value() == "Time"
    assert await zebra.pc.pulse_source.get_value() == "Position"


def test_get_zebra_settings_for_extruder_pumpprobe():
    width, step = get_zebra_settings_for_extruder(0.01, 0.005, 0.001)
    assert round(width, 3) == 0.016
    assert round(step, 3) == 0.026


async def test_setup_zebra_for_quickshot(zebra: Zebra, RE):
    RE(setup_zebra_for_quickshot_plan(zebra, exp_time=0.001, num_images=10, wait=True))
    assert await zebra.pc.arm_source.get_value() == "Soft"
    assert await zebra.pc.gate_start.get_value() == 1.0
    assert await zebra.pc.gate_input.get_value() == zebra.mapping.sources.SOFT_IN2


async def test_setup_zebra_for_extruder_pp_eiger_collection(zebra: Zebra, RE):
    inputs_list = (0.01, 10, 0.005, 0.001)
    # With eiger
    RE(
        setup_zebra_for_extruder_with_pump_probe_plan(
            zebra, "eiger", *inputs_list, wait=True
        )
    )
    assert await zebra.output.out_pvs[1].get_value() == zebra.mapping.sources.AND4
    assert await zebra.output.out_pvs[2].get_value() == zebra.mapping.sources.AND3

    assert await zebra.inputs.soft_in_1.get_value() == "No"
    assert (
        await zebra.logic_gates.and_gates[3].sources[1].get_value()
        == zebra.mapping.sources.SOFT_IN2
    )
    assert await zebra.pc.num_gates.get_value() == 10


async def test_setup_zebra_for_extruder_pp_pilatus_collection(zebra: Zebra, RE):
    inputs_list = (0.01, 10, 0.005, 0.001)
    # With pilatus
    RE(
        setup_zebra_for_extruder_with_pump_probe_plan(
            zebra, "pilatus", *inputs_list, wait=True
        )
    )
    # Check that SOFT_IN:B0 gets disabled
    assert await zebra.output.out_pvs[1].get_value() == zebra.mapping.sources.AND3

    assert await zebra.pc.gate_start.get_value() == 1.0
    assert await zebra.output.pulse_1.delay.get_value() == 0.0
    assert await zebra.output.pulse_2.delay.get_value() == 0.001


async def test_setup_zebra_for_fastchip(zebra: Zebra, RE):
    num_gates = 400
    num_exposures = 2
    exposure_time_s = 0.001
    # With Eiger
    RE(
        setup_zebra_for_fastchip_plan(
            zebra, "eiger", num_gates, num_exposures, exposure_time_s, wait=True
        )
    )
    # Check that SOFT_IN:B0 gets disabled
    assert await zebra.output.out_pvs[1].get_value() == zebra.mapping.sources.AND3

    # Check ttl out1 is set to AND3
    assert await zebra.output.out_pvs[1].get_value() == zebra.mapping.sources.AND3
    assert await zebra.pc.num_gates.get_value() == num_gates
    assert await zebra.pc.pulse_max.get_value() == num_exposures
    assert await zebra.pc.pulse_width.get_value() == exposure_time_s - 0.0001

    # With Pilatus
    RE(
        setup_zebra_for_fastchip_plan(
            zebra, "pilatus", num_gates, num_exposures, exposure_time_s, wait=True
        )
    )
    # Check ttl out2 is set to AND3
    assert await zebra.output.out_pvs[2].get_value() == zebra.mapping.sources.AND3

    assert await zebra.pc.pulse_start.get_value() == 0.0
    assert await zebra.pc.pulse_width.get_value() == exposure_time_s / 2
    assert await zebra.pc.pulse_step.get_value() == exposure_time_s + 0.0001


async def test_open_fast_shutter_at_each_position_plan(zebra: Zebra, RE):
    num_exposures = 2
    exposure_time_s = 0.001

    RE(open_fast_shutter_at_each_position_plan(zebra, num_exposures, exposure_time_s))

    # Check output Pulse2 is set
    assert await zebra.output.pulse_2.input.get_value() == zebra.mapping.sources.PC_GATE
    assert await zebra.output.pulse_2.delay.get_value() == 0.0
    expected_pulse_width = num_exposures * exposure_time_s + 0.05
    assert await zebra.output.pulse_2.width.get_value() == pytest.approx(
        expected_pulse_width, abs=1e-3
    )

    assert await zebra.output.out_pvs[4].get_value() == zebra.mapping.sources.PULSE2


async def test_reset_pc_gate_and_pulse(zebra: Zebra, RE):
    RE(reset_pc_gate_and_pulse(zebra))

    assert await zebra.pc.gate_start.get_value() == 0
    assert await zebra.pc.pulse_width.get_value() == 0
    assert await zebra.pc.pulse_step.get_value() == 0


async def test_reset_output_panel(zebra: Zebra, RE):
    RE(reset_output_panel(zebra))

    assert await zebra.output.out_pvs[2].get_value() == zebra.mapping.sources.PC_GATE
    assert await zebra.output.out_pvs[4].get_value() == zebra.mapping.sources.OR1
    assert (
        await zebra.output.pulse_1.input.get_value() == zebra.mapping.sources.DISCONNECT
    )
    assert (
        await zebra.output.pulse_2.input.get_value() == zebra.mapping.sources.DISCONNECT
    )


async def test_zebra_return_to_normal(zebra: Zebra, RE):
    RE(zebra_return_to_normal_plan(zebra, wait=True))

    assert await zebra.pc.reset.get_value() == 1
    assert await zebra.pc.gate_source.get_value() == "Position"
    assert await zebra.pc.pulse_source.get_value() == "Position"
    assert await zebra.pc.gate_trigger.get_value() == "Enc2"
    assert await zebra.pc.gate_start.get_value() == 0

    assert await zebra.output.out_pvs[3].get_value() == zebra.mapping.sources.DISCONNECT
    assert (
        await zebra.output.pulse_1.input.get_value() == zebra.mapping.sources.DISCONNECT
    )


async def test_reset_zebra_plan(zebra: Zebra, RE):
    RE(reset_zebra_when_collection_done_plan(zebra))

    assert await zebra.inputs.soft_in_2.get_value() == "No"
    assert await zebra.pc.is_armed() is False



from typing import Protocol, runtime_checkable

from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.xbpm_feedback import XBPMFeedback


@runtime_checkable
class XBPMPauseDevices(Protocol):
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator



import os
from logging import FileHandler
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from dodal.log import LOGGER as dodal_logger
from dodal.log import set_up_all_logging_handlers

import mx_bluesky.common.utils.log as log
from mx_bluesky.common.external_interaction.callbacks.common.log_uid_tag_callback import (
    LogUidTaggingCallback,
)

from ....conftest import clear_log_handlers

TEST_GRAYLOG_PORT = 5555


@pytest.fixture(scope="function")
def clear_and_mock_loggers():
    clear_log_handlers([*log.ALL_LOGGERS, dodal_logger])
    mock_open_with_tell = MagicMock()
    mock_open_with_tell.tell.return_value = 0
    with (
        patch("dodal.log.logging.FileHandler._open", mock_open_with_tell),
        patch("dodal.log.GELFTCPHandler.emit") as graylog_emit,
        patch("dodal.log.TimedRotatingFileHandler.emit") as filehandler_emit,
    ):
        graylog_emit.reset_mock()
        filehandler_emit.reset_mock()
        yield filehandler_emit, graylog_emit
    clear_log_handlers([*log.ALL_LOGGERS, dodal_logger])


@pytest.mark.skip_log_setup
def test_no_env_variable_sets_correct_file_handler(
    clear_and_mock_loggers,
) -> None:
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)
    file_handlers: FileHandler = next(
        filter(lambda h: isinstance(h, FileHandler), dodal_logger.handlers)  # type: ignore
    )

    assert file_handlers.baseFilename.endswith("/tmp/logs/bluesky/hyperion.log")


@pytest.mark.skip_log_setup
@patch("dodal.log.Path.mkdir", autospec=True)
@patch.dict(
    os.environ, {"LOG_DIR": "./dls_sw/s03/logs/bluesky"}
)  # Note we use a relative path here so it works in CI
def test_set_env_variable_sets_correct_file_handler(
    mock_dir,
    clear_and_mock_loggers,
) -> None:
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)

    file_handlers: FileHandler = next(
        filter(lambda h: isinstance(h, FileHandler), dodal_logger.handlers)  # type: ignore
    )

    assert file_handlers.baseFilename.endswith("/dls_sw/s03/logs/bluesky/hyperion.log")


@pytest.mark.skip_log_setup
def test_messages_logged_from_dodal_and_hyperion_contain_dcgid(
    clear_and_mock_loggers,
):
    _, mock_GELFTCPHandler_emit = clear_and_mock_loggers
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)

    log.set_dcgid_tag(100)

    logger = log.LOGGER
    logger.info("test_hyperion")
    dodal_logger.info("test_dodal")

    graylog_calls = mock_GELFTCPHandler_emit.mock_calls[1:]

    dc_group_id_correct = [c.args[0].dc_group_id == 100 for c in graylog_calls]
    assert all(dc_group_id_correct)


@pytest.mark.skip_log_setup
def test_messages_are_tagged_with_run_uid(clear_and_mock_loggers, RE):
    _, mock_GELFTCPHandler_emit = clear_and_mock_loggers
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)

    RE.subscribe(LogUidTaggingCallback())
    test_run_uid = None
    logger = log.LOGGER

    @bpp.run_decorator()
    def test_plan():
        yield from bps.sleep(0)
        assert log.tag_filter.run_uid is not None
        nonlocal test_run_uid
        test_run_uid = log.tag_filter.run_uid
        logger.info("test_hyperion")
        logger.info("test_hyperion")
        yield from bps.sleep(0)

    assert log.tag_filter.run_uid is None
    RE(test_plan())
    assert log.tag_filter.run_uid is None

    graylog_calls_in_plan = [
        c.args[0]
        for c in mock_GELFTCPHandler_emit.mock_calls
        if c.args[0].msg == "test_hyperion"
    ]

    assert len(graylog_calls_in_plan) == 2

    dc_group_id_correct = [
        record.run_uid == test_run_uid for record in graylog_calls_in_plan
    ]
    assert all(dc_group_id_correct)


@pytest.mark.skip_log_setup
def test_messages_logged_from_dodal_and_hyperion_get_sent_to_graylog_and_file(
    clear_and_mock_loggers,
):
    mock_filehandler_emit, mock_GELFTCPHandler_emit = clear_and_mock_loggers
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)
    logger = log.LOGGER
    logger.info("test MX_Bluesky")
    dodal_logger.info("test_dodal")

    filehandler_calls = mock_filehandler_emit.mock_calls
    graylog_calls = mock_GELFTCPHandler_emit.mock_calls

    assert len(filehandler_calls) >= 2
    assert len(graylog_calls) >= 2

    for handler in [filehandler_calls, graylog_calls]:
        handler_names = [c.args[0].name for c in handler]
        handler_messages = [c.args[0].message for c in handler]
        assert "MX-Bluesky" in handler_names
        assert "Dodal" in handler_names
        assert "test MX_Bluesky" in handler_messages
        assert "test_dodal" in handler_messages


@pytest.mark.skip_log_setup
def test_callback_loggers_log_to_own_files(
    clear_and_mock_loggers,
):
    mock_filehandler_emit, mock_GELFTCPHandler_emit = clear_and_mock_loggers
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)

    hyperion_logger = log.LOGGER
    ISPYB_ZOCALO_CALLBACK_LOGGER = log.ISPYB_ZOCALO_CALLBACK_LOGGER
    nexus_logger = log.NEXUS_LOGGER
    for logger in [ISPYB_ZOCALO_CALLBACK_LOGGER, nexus_logger]:
        set_up_all_logging_handlers(
            logger, log._get_logging_dir(), logger.name, True, 10000
        )

    hyperion_logger.info("test_hyperion")
    ISPYB_ZOCALO_CALLBACK_LOGGER.info("test_ispyb")
    nexus_logger.info("test_nexus")

    total_filehandler_calls = mock_filehandler_emit.mock_calls
    total_graylog_calls = mock_GELFTCPHandler_emit.mock_calls

    assert len(total_filehandler_calls) == len(total_graylog_calls)

    hyperion_filehandler = next(
        filter(lambda h: isinstance(h, TimedRotatingFileHandler), dodal_logger.handlers)  # type: ignore
    )
    ispyb_filehandler = next(
        filter(
            lambda h: isinstance(h, TimedRotatingFileHandler),
            ISPYB_ZOCALO_CALLBACK_LOGGER.handlers,
        )  # type: ignore
    )
    nexus_filehandler = next(
        filter(lambda h: isinstance(h, TimedRotatingFileHandler), nexus_logger.handlers)  # type: ignore
    )
    assert nexus_filehandler.baseFilename != hyperion_filehandler.baseFilename  # type: ignore
    assert ispyb_filehandler.baseFilename != hyperion_filehandler.baseFilename  # type: ignore
    assert ispyb_filehandler.baseFilename != nexus_filehandler.baseFilename  # type: ignore


@pytest.mark.skip_log_setup
def test_log_writes_debug_file_on_error(clear_and_mock_loggers):
    mock_filehandler_emit, _ = clear_and_mock_loggers
    log.do_default_logging_setup("hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True)
    log.LOGGER.debug("debug_message_1")
    log.LOGGER.debug("debug_message_2")
    mock_filehandler_emit.assert_not_called()
    log.LOGGER.error("error happens")
    assert len(mock_filehandler_emit.mock_calls) == 4
    messages = [call.args[0].message for call in mock_filehandler_emit.mock_calls]
    assert "debug_message_1" in messages
    assert "debug_message_2" in messages
    assert "error happens" in messages


@patch("mx_bluesky.common.utils.log.Path.mkdir")
def test_get_logging_dir_uses_env_var(mock_mkdir: MagicMock):
    def mock_get_log_dir(variable: str):
        if variable == "LOG_DIR":
            return "test_dir"
        else:
            return "other_dir"

    with patch(
        "mx_bluesky.common.utils.log.environ.get",
        side_effect=lambda var: mock_get_log_dir(var),
    ):
        assert log._get_logging_dir() == Path("test_dir")
        mock_mkdir.assert_called_once()


@patch("mx_bluesky.common.utils.log.Path.mkdir")
def test_get_logging_dir_uses_beamline_if_no_dir_env_var(mock_mkdir: MagicMock):
    def mock_get_log_dir(variable: str):
        if variable == "LOG_DIR":
            return None
        elif variable == "BEAMLINE":
            return "test"

    with patch(
        "mx_bluesky.common.utils.log.environ.get",
        side_effect=lambda var: mock_get_log_dir(var),
    ):
        assert log._get_logging_dir() == Path("/dls_sw/test/logs/bluesky/")
        mock_mkdir.assert_called_once()


@patch("mx_bluesky.common.utils.log.Path.mkdir")
def test_get_logging_dir_uses_tmp_if_no_env_var(mock_mkdir: MagicMock):
    assert log._get_logging_dir() == Path("/tmp/logs/bluesky")
    mock_mkdir.assert_called_once()


@pytest.mark.skip_log_setup
@patch("mx_bluesky.common.utils.log.Path.mkdir")
@patch(
    "mx_bluesky.common.utils.log.integrate_bluesky_and_ophyd_logging",
)
def test_default_logging_setup_integrate_logs_flag(
    mock_integrate_logs: MagicMock, mock_mkdir
):
    log.do_default_logging_setup(
        "hyperion.log", TEST_GRAYLOG_PORT, dev_mode=True, integrate_all_logs=False
    )
    mock_integrate_logs.assert_not_called()



import re
import subprocess
from os import environ
from unittest.mock import patch

import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine

from mx_bluesky.common.plans.read_hardware import (
    standard_read_hardware_during_collection,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ....conftest import extract_metafile, raw_params_from_file

DOCKER = environ.get("DOCKER", "docker")


@pytest.fixture
def test_params(tmpdir):
    param_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
    )
    params = RotationScan(**param_dict)
    params.demand_energy_ev = 12700
    params.scan_width_deg = 360
    params.storage_directory = "tests/test_data"
    params.x_start_um = 0
    params.y_start_um = 0
    params.z_start_um = 0
    params.exposure_time_s = 0.004
    return params


@pytest.mark.parametrize(
    "test_data_directory, prefix, reference_file",
    [
        (
            "tests/test_data/nexus_files/rotation",
            "ins_8_5",
            "ins_8_5_expected_output.txt",
        ),
        (
            "tests/test_data/nexus_files/rotation_unicode_metafile",
            "ins_8_5",
            "ins_8_5_expected_output.txt",
        ),
    ],
)
@pytest.mark.system_test
def test_rotation_nexgen(
    test_params: RotationScan,
    tmpdir,
    fake_create_rotation_devices: RotationScanComposite,
    test_data_directory,
    prefix,
    reference_file,
):
    meta_file = f"{prefix}_meta.h5.gz"
    test_params.file_name = prefix
    test_params.storage_directory = f"{tmpdir}"
    run_number = test_params.detector_params.run_number

    extract_metafile(
        f"{test_data_directory}/{meta_file}", f"{tmpdir}/{prefix}_{run_number}_meta.h5"
    )

    fake_create_rotation_devices.eiger.bit_depth.sim_put(32)  # type: ignore

    RE = RunEngine({})

    with patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.get_start_and_predicted_end_time",
        return_value=("test_time", "test_time"),
    ):
        RE(
            _fake_rotation_scan(
                test_params, RotationNexusFileCallback(), fake_create_rotation_devices
            )
        )

    master_file = f"{tmpdir}/{prefix}_{run_number}_master.h5"
    _check_nexgen_output_passes_imginfo(
        master_file, f"{test_data_directory}/{reference_file}"
    )


FILE_PATTERN = re.compile("^ ################# File = (.*)")

HEADER_PATTERN = re.compile("^ ===== Header information:")

DATE_PATTERN = re.compile("^ date                                = (.*)")


def _check_nexgen_output_passes_imginfo(test_file, reference_file):
    stdout, stderr = _run_imginfo(test_file)
    assert stderr == ""
    it_actual_lines = iter(stdout.split("\n"))
    i = 0
    try:
        with open(reference_file) as f:
            while True:
                i += 1
                expected_line = f.readline().rstrip("\n")
                actual_line = next(it_actual_lines)
                if FILE_PATTERN.match(actual_line):
                    continue
                if HEADER_PATTERN.match(actual_line):
                    break
                assert actual_line == expected_line, (
                    f"Header line {i} didn't match contents of {reference_file}: {actual_line} <-> {expected_line}"
                )

            while True:
                i += 1
                expected_line = f.readline().rstrip("\n")
                actual_line = next(it_actual_lines)
                if DATE_PATTERN.match(actual_line):
                    continue
                assert actual_line == expected_line, (
                    f"Header line {i} didn't match contents of {reference_file}: {actual_line} <-> {expected_line}"
                )

    except StopIteration:
        pass

        # assert stdout == expected


def _run_imginfo(filename):
    process = subprocess.run(
        # This file is provided in the system test docker image
        ["/usr/local/bin/imginfo", filename],
        text=True,
        capture_output=True,
    )
    assert process.returncode != 2, "imginfo is not available"
    assert process.returncode == 0, (
        f"imginfo failed with returncode {process.returncode}"
    )

    return process.stdout, process.stderr


def _fake_rotation_scan(
    parameters: RotationScan,
    subscription: RotationNexusFileCallback,
    rotation_devices: RotationScanComposite,
):
    @bpp.subs_decorator(subscription)
    @bpp.set_run_key_decorator("rotation_scan_with_cleanup_and_subs")
    @bpp.run_decorator(  # attach experiment metadata to the start document
        md={
            "subplan_name": CONST.PLAN.ROTATION_OUTER,
            "mx_bluesky_parameters": parameters.model_dump_json(),
            "activate_callbacks": "RotationNexusFileCallback",
        }
    )
    def plan():
        yield from standard_read_hardware_during_collection(
            rotation_devices.aperture_scatterguard,
            rotation_devices.attenuator,
            rotation_devices.flux,
            rotation_devices.dcm,
            rotation_devices.eiger,
        )

    return plan()



import os
from collections.abc import Callable, Generator, Sequence
from functools import partial
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import ispyb.sqlalchemy
import numpy
import pytest
import pytest_asyncio
from dodal.beamlines import i03
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.dcm import DCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.flux import Flux
from dodal.devices.i03.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from ispyb.sqlalchemy import (
    BLSample,
    DataCollection,
    DataCollectionGroup,
    GridInfo,
    Position,
)
from ophyd.sim import NullStatus
from ophyd_async.core import AsyncStatus
from ophyd_async.testing import callback_on_mock_put, set_mock_value
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from workflows.recipe import RecipeWrapper

from mx_bluesky.common.external_interaction.ispyb.ispyb_store import StoreInIspyb
from mx_bluesky.common.utils.utils import convert_angstrom_to_eV
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ....conftest import (
    TEST_RESULT_MEDIUM,
    fake_read,
    pin_tip_edge_data,
    raw_params_from_file,
)


def get_current_datacollection_comment(Session: Callable, dcid: int) -> str:
    """Read the 'comments' field from the given datacollection id's ISPyB entry.
    Returns an empty string if the comment is not yet initialised.
    """
    try:
        with Session() as session:
            query = session.query(DataCollection).filter(
                DataCollection.dataCollectionId == dcid
            )
            current_comment: str = query.first().comments
    except Exception:
        current_comment = ""
    return current_comment


def get_datacollections(Session: Callable, dcg_id: int) -> Sequence[int]:
    with Session.begin() as session:  # type: ignore
        query = session.query(DataCollection.dataCollectionId).filter(
            DataCollection.dataCollectionGroupId == dcg_id
        )
        return [row[0] for row in query.all()]


def get_current_datacollection_attribute(
    Session: Callable, dcid: int, attr: str
) -> str:
    """Read the specified field 'attr' from the given datacollection id's ISPyB entry.
    Returns an empty string if the attribute is not found.
    """
    try:
        with Session() as session:
            query = session.query(DataCollection).filter(
                DataCollection.dataCollectionId == dcid
            )
            first_result = query.first()
            data: str = getattr(first_result, attr)
    except Exception:
        data = ""
    return data


def get_current_datacollection_grid_attribute(
    Session: Callable, grid_id: int, attr: str
) -> Any:
    with Session() as session:
        query = session.query(GridInfo).filter(GridInfo.gridInfoId == grid_id)
        first_result = query.first()
        return getattr(first_result, attr)


def get_current_position_attribute(
    Session: Callable, position_id: int, attr: str
) -> Any:
    with Session() as session:
        query = session.query(Position).filter(Position.positionId == position_id)
        first_result = query.first()
        if first_result is None:
            return None
        return getattr(first_result, attr)


def get_current_datacollectiongroup_attribute(
    Session: Callable, dcg_id: int, attr: str
):
    with Session() as session:
        query = session.query(DataCollectionGroup).filter(
            DataCollectionGroup.dataCollectionGroupId == dcg_id
        )
        first_result = query.first()
        return getattr(first_result, attr)


def get_blsample(Session: Callable, bl_sample_id: int) -> BLSample:
    with Session() as session:
        query = session.query(BLSample).filter(BLSample.blSampleId == bl_sample_id)
        return query.first()


@pytest.fixture
def sqlalchemy_sessionmaker() -> sessionmaker:
    url = ispyb.sqlalchemy.url(CONST.SIM.DEV_ISPYB_DATABASE_CFG)
    engine = create_engine(url, connect_args={"use_pure": True})
    return sessionmaker(engine)


@pytest.fixture
def fetch_comment(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollection_comment, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_ids_for_group_id(
    sqlalchemy_sessionmaker,
) -> Callable[[int], Sequence]:
    return partial(get_datacollections, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollection_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_grid_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollection_grid_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_position_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_position_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollectiongroup_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollectiongroup_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_blsample(sqlalchemy_sessionmaker) -> Callable[[int], BLSample]:
    return partial(get_blsample, sqlalchemy_sessionmaker)


@pytest.fixture
def dummy_params():
    dummy_params = HyperionSpecifiedThreeDGridScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/test_gridscan_param_defaults.json"
        )
    )
    dummy_params.visit = os.environ.get("ST_VISIT", "cm31105-5")
    dummy_params.sample_id = int(os.environ.get("ST_SAMPLE_ID", dummy_params.sample_id))
    return dummy_params


@pytest.fixture
def dummy_ispyb(
    use_dev_ispyb_unless_overridden_by_environment, dummy_params
) -> StoreInIspyb:
    return StoreInIspyb(use_dev_ispyb_unless_overridden_by_environment)


@pytest.fixture
def dummy_ispyb_3d(
    use_dev_ispyb_unless_overridden_by_environment, dummy_params
) -> StoreInIspyb:
    return StoreInIspyb(use_dev_ispyb_unless_overridden_by_environment)


@pytest.fixture
def zocalo_env():
    os.environ["ZOCALO_CONFIG"] = "/dls_sw/apps/zocalo/live/configuration.yaml"


@pytest_asyncio.fixture
async def zocalo_for_fake_zocalo(zocalo_env) -> ZocaloResults:
    """
    This attempts to connect to a fake zocalo via rabbitmq
    """
    zd = ZocaloResults()
    zd.timeout_s = 10
    await zd.connect()
    return zd


@pytest.fixture
def zocalo_for_system_test() -> Generator[ZocaloResults, None, None]:
    zocalo = i03.zocalo(connect_immediately=True, mock=True)
    old_zocalo_trigger = zocalo.trigger
    zocalo.my_zocalo_result = TEST_RESULT_MEDIUM

    @AsyncStatus.wrap
    async def mock_zocalo_complete():
        fake_recipe_wrapper = MagicMock(spec=RecipeWrapper)
        fake_recipe_wrapper.recipe_step = {"parameters": {"dcid": 1234, "dcgid": 123}}
        message = {
            "results": zocalo.my_zocalo_result  # type: ignore
        }
        header = {}
        zocalo.my_callback(fake_recipe_wrapper, header, message)  # type: ignore
        await old_zocalo_trigger()

    def mock_worfklow_subscribe(transport, channel, callback, **kwargs):
        if channel == "xrc.i03":
            zocalo.my_callback = callback

    with (
        patch("dodal.devices.zocalo.zocalo_results.workflows") as workflows,
        patch("dodal.devices.zocalo.zocalo_results._get_zocalo_connection"),
    ):
        workflows.recipe.wrap_subscribe.side_effect = mock_worfklow_subscribe
        with patch.object(zocalo, "trigger", side_effect=mock_zocalo_complete):
            yield zocalo


@pytest.fixture
def grid_detect_then_xray_centre_composite(
    fast_grid_scan,
    backlight,
    beamstop_i03,
    smargon,
    undulator_for_system_test,
    synchrotron,
    s4_slit_gaps,
    attenuator,
    xbpm_feedback,
    detector_motion,
    zocalo_for_system_test,
    aperture_scatterguard,
    zebra,
    eiger,
    robot,
    oav_for_system_test,
    dcm,
    flux,
    ophyd_pin_tip_detection,
    sample_shutter,
    panda,
    panda_fast_grid_scan,
):
    composite = GridDetectThenXRayCentreComposite(
        zebra_fast_grid_scan=fast_grid_scan,
        pin_tip_detection=ophyd_pin_tip_detection,
        backlight=backlight,
        beamstop=beamstop_i03,
        panda_fast_grid_scan=panda_fast_grid_scan,
        smargon=smargon,
        undulator=undulator_for_system_test,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        attenuator=attenuator,
        xbpm_feedback=xbpm_feedback,
        detector_motion=detector_motion,
        zocalo=zocalo_for_system_test,
        aperture_scatterguard=aperture_scatterguard,
        zebra=zebra,
        eiger=eiger,
        panda=panda,
        robot=robot,
        oav=oav_for_system_test,
        dcm=dcm,
        flux=flux,
        sample_shutter=sample_shutter,
    )

    @AsyncStatus.wrap
    async def mock_pin_tip_detect():
        tip_x_px, tip_y_px, top_edge_array, bottom_edge_array = pin_tip_edge_data()
        set_mock_value(
            ophyd_pin_tip_detection.triggered_top_edge,
            top_edge_array,
        )

        set_mock_value(
            ophyd_pin_tip_detection.triggered_bottom_edge,
            bottom_edge_array,
        )
        set_mock_value(
            zocalo_for_system_test.bounding_box,
            numpy.array([[10, 10, 10]], dtype=numpy.uint64),
        )
        set_mock_value(
            ophyd_pin_tip_detection.triggered_tip, numpy.array([tip_x_px, tip_y_px])
        )

    with (
        patch.object(eiger, "wait_on_arming_if_started"),
        # xsize, ysize will always be wrong since computed as 0 before we get here
        # patch up load_microns_per_pixel connect to receive non-zero values
        patch.object(
            ophyd_pin_tip_detection, "trigger", side_effect=mock_pin_tip_detect
        ),
        patch.object(fast_grid_scan, "kickoff", return_value=NullStatus()),
        patch.object(fast_grid_scan, "complete", return_value=NullStatus()),
    ):
        yield composite


@pytest.fixture
def fgs_composite_for_fake_zocalo(
    fake_fgs_composite: HyperionFlyScanXRayCentreComposite,
    zocalo_for_fake_zocalo: ZocaloResults,
    done_status: NullStatus,
) -> HyperionFlyScanXRayCentreComposite:
    set_mock_value(fake_fgs_composite.aperture_scatterguard.aperture.z.user_setpoint, 2)
    fake_fgs_composite.eiger.unstage = MagicMock(return_value=done_status)  # type: ignore
    fake_fgs_composite.smargon.stub_offsets.set = MagicMock(return_value=done_status)  # type: ignore
    callback_on_mock_put(
        fake_fgs_composite.zebra_fast_grid_scan.run_cmd,
        lambda *args, **kwargs: set_mock_value(
            fake_fgs_composite.zebra_fast_grid_scan.status, 1
        ),
    )
    fake_fgs_composite.zebra_fast_grid_scan.complete = MagicMock(
        return_value=NullStatus()
    )
    fake_fgs_composite.zocalo = zocalo_for_fake_zocalo
    return fake_fgs_composite


@pytest.fixture
def pin_tip_no_pin_found(ophyd_pin_tip_detection):
    @AsyncStatus.wrap
    async def no_pin_tip_found():
        set_mock_value(
            ophyd_pin_tip_detection.triggered_tip, PinTipDetection.INVALID_POSITION
        )

        set_mock_value(
            ophyd_pin_tip_detection.triggered_top_edge,
            numpy.array([]),
        )

        set_mock_value(
            ophyd_pin_tip_detection.triggered_bottom_edge,
            numpy.array([]),
        )

    with patch.object(ophyd_pin_tip_detection, "trigger", side_effect=no_pin_tip_found):
        yield ophyd_pin_tip_detection


@pytest.fixture
def params_for_rotation_scan(test_rotation_params: RotationScan) -> RotationScan:
    test_rotation_params.rotation_increment_deg = 0.27
    test_rotation_params.exposure_time_s = 0.023
    test_rotation_params.detector_params.expected_energy_ev = 0.71
    test_rotation_params.visit = os.environ.get("ST_VISIT", "cm31105-4")
    test_rotation_params.sample_id = int(os.environ.get("ST_SAMPLE_ID", 123456))
    return test_rotation_params


@pytest.fixture
def composite_for_rotation_scan(
    beamstop_i03: Beamstop,
    eiger: EigerDetector,
    smargon: Smargon,
    zebra: Zebra,
    detector_motion: DetectorMotion,
    backlight: Backlight,
    attenuator: BinaryFilterAttenuator,
    flux: Flux,
    undulator_for_system_test: Undulator,
    aperture_scatterguard: ApertureScatterguard,
    synchrotron: Synchrotron,
    s4_slit_gaps: S4SlitGaps,
    dcm: DCM,
    robot: BartRobot,
    oav_for_system_test: OAV,
    sample_shutter: ZebraShutter,
    xbpm_feedback: XBPMFeedback,
):
    set_mock_value(smargon.omega.max_velocity, 131)
    oav_for_system_test.zoom_controller.level.describe = AsyncMock(
        return_value={"level": {"choices": ["1.0x", "5.0x", "7.5x"]}}
    )

    fake_create_rotation_devices = RotationScanComposite(
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_i03,
        dcm=dcm,
        detector_motion=detector_motion,
        eiger=eiger,
        flux=flux,
        smargon=smargon,
        undulator=undulator_for_system_test,
        aperture_scatterguard=aperture_scatterguard,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        zebra=zebra,
        robot=robot,
        oav=oav_for_system_test,
        sample_shutter=sample_shutter,
        xbpm_feedback=xbpm_feedback,
    )

    energy_ev = convert_angstrom_to_eV(0.71)
    set_mock_value(
        fake_create_rotation_devices.dcm.energy_in_kev.user_readback,
        energy_ev / 1000,  # pyright: ignore
    )
    set_mock_value(
        fake_create_rotation_devices.synchrotron.synchrotron_mode,
        SynchrotronMode.USER,
    )
    set_mock_value(
        fake_create_rotation_devices.synchrotron.top_up_start_countdown,
        -1,
    )
    set_mock_value(fake_create_rotation_devices.s4_slit_gaps.xgap.user_readback, 0.123)
    set_mock_value(fake_create_rotation_devices.s4_slit_gaps.ygap.user_readback, 0.234)

    with (
        patch("bluesky.preprocessors.__read_and_stash_a_motor", fake_read),
        patch("bluesky.plan_stubs.wait"),
    ):
        yield fake_create_rotation_devices



from abc import ABC, abstractmethod
from functools import cache

from daq_config_server.client import ConfigServer
from pydantic import BaseModel, Field, model_validator


class FeatureFlags(BaseModel, ABC):
    """Abstract class to use ConfigServer to toggle features for an experiment

    A module wanting to use FeatureFlags should inherit this class, add boolean features
    as attributes, and implement a get_config_server method, which returns a cached creation of
    ConfigServer. See HyperionFeatureFlags for an example

    Values supplied upon class instantiation will always take priority over the config server. If connection to the server cannot
    be made AND values were not supplied, attributes will use their default values
    """

    # Feature values supplied at construction will override values from the config server
    overriden_features: dict = Field(default_factory=dict, exclude=True)

    @staticmethod
    @cache
    @abstractmethod
    def get_config_server() -> ConfigServer: ...

    @model_validator(mode="before")
    @classmethod
    def mark_overridden_features(cls, values):
        assert isinstance(values, dict)
        values["overriden_features"] = values.copy()
        cls._validate_overridden_features(values)
        return values

    @classmethod
    def _validate_overridden_features(cls, values: dict):
        """Validates overridden features to ensure they are defined in the model fields."""
        defined_fields = cls.model_fields.keys()
        invalid_features = [key for key in values.keys() if key not in defined_fields]

        if invalid_features:
            message = f"Invalid feature toggle(s) supplied: {invalid_features}. "
            raise ValueError(message)

    def _get_flags(self):
        flags = type(self).get_config_server().best_effort_get_all_feature_flags()
        return {f: flags[f] for f in flags if f in self.model_fields.keys()}

    def update_self_from_server(self):
        """Used to update the feature flags from the server during a plan. Where there are flags which were explicitly set from externally supplied parameters, these values will be used instead."""
        for flag, value in self._get_flags().items():
            updated_value = (
                value
                if flag not in self.overriden_features.keys()
                else self.overriden_features[flag]
            )
            setattr(self, flag, updated_value)



from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from event_model import RunStart

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
)
from mx_bluesky.common.parameters.components import IspybExperimentType
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import rotation_scan
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_rotation_callbacks,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from .....conftest import raw_params_from_file


@pytest.fixture
def params():
    return RotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
        )
    )


def activate_callbacks(cbs: tuple[RotationNexusFileCallback, RotationISPyBCallback]):
    cbs[1].active = True
    cbs[0].active = True


@pytest.fixture
def do_rotation_scan(
    params: RotationScan, fake_create_rotation_devices, oav_parameters_for_rotation
):
    return rotation_scan(
        fake_create_rotation_devices, params, oav_parameters_for_rotation
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback.NexusWriter",
    autospec=True,
)
def test_nexus_handler_gets_documents_in_plan(
    nexus_writer: MagicMock,
    do_rotation_scan,
    RE: RunEngine,
):
    nexus_writer.return_value.data_filename = "test_full_filename"
    nexus_callback, _ = create_rotation_callbacks()
    activate_callbacks((nexus_callback, _))
    nexus_callback.activity_gated_start = MagicMock(
        side_effect=nexus_callback.activity_gated_start
    )
    RE.subscribe(nexus_callback)
    RE(do_rotation_scan)

    subplans = []
    for call in nexus_callback.activity_gated_start.call_args_list:  #  type: ignore
        subplans.append(call.args[0].get("subplan_name"))

    assert CONST.PLAN.ROTATION_OUTER in subplans


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback.NexusWriter",
    autospec=True,
)
def test_nexus_handler_only_writes_once(
    nexus_writer: MagicMock, RE: RunEngine, do_rotation_scan
):
    nexus_writer.return_value.data_filename = "test_full_filename"
    cb = RotationNexusFileCallback()
    cb.active = True
    RE.subscribe(cb)
    RE(do_rotation_scan)
    nexus_writer.assert_called_once()
    assert cb.writer is not None
    cb.writer.create_nexus_file.assert_called_once()  # type: ignore


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
def test_ispyb_handler_receives_two_stops_but_only_ends_deposition_on_inner_one(
    ispyb_store, zocalo, RE: RunEngine, do_rotation_scan
):
    _, ispyb_callback = create_rotation_callbacks()
    ispyb_callback.emit_cb = None
    ispyb_callback.activity_gated_start = MagicMock(
        autospec=True, side_effect=ispyb_callback.activity_gated_start
    )
    ispyb_callback.activity_gated_stop = MagicMock(
        autospec=True, side_effect=ispyb_callback.activity_gated_stop
    )

    parent_mock = MagicMock()
    parent_mock.attach_mock(ispyb_store.return_value.end_deposition, "end_deposition")
    parent_mock.attach_mock(ispyb_callback.activity_gated_stop, "callback_stopped")

    RE.subscribe(ispyb_callback)
    RE(do_rotation_scan)

    assert ispyb_callback.activity_gated_stop.call_count == 2
    assert parent_mock.method_calls[1][0] == "end_deposition"


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb",
    autospec=True,
)
def test_ispyb_reuses_dcgid_on_same_sampleID(
    rotation_ispyb: MagicMock,
    RE: RunEngine,
    params: RotationScan,
    fake_create_rotation_devices,
    oav_parameters_for_rotation,
):
    ispyb_cb = RotationISPyBCallback()
    ispyb_cb.active = True
    ispyb_ids = IspybIds(data_collection_group_id=23, data_collection_ids=(45,))
    rotation_ispyb.return_value.begin_deposition.return_value = ispyb_ids

    test_cases = zip(
        [123, 123, 123, 456, 123],
        [False, True, True, False, False],
        strict=False,
    )

    last_dcgid = None

    RE.subscribe(ispyb_cb)

    for sample_id, same_dcgid in test_cases:
        params.sample_id = sample_id

        RE(
            rotation_scan(
                fake_create_rotation_devices, params, oav_parameters_for_rotation
            )
        )

        begin_deposition_scan_data: ScanDataInfo = (
            rotation_ispyb.return_value.begin_deposition.call_args.args[1][0]
        )
        if same_dcgid:
            assert begin_deposition_scan_data.data_collection_info.parent_id is not None
            assert (
                begin_deposition_scan_data.data_collection_info.parent_id is last_dcgid
            )
        else:
            assert begin_deposition_scan_data.data_collection_info.parent_id is None

        last_dcgid = ispyb_cb.ispyb_ids.data_collection_group_id


n_images_store_id = [
    (123, False),
    (3600, True),
    (1800, True),
    (150, False),
    (500, True),
    (201, True),
    (1, False),
    (2000, True),
    (2000, True),
    (2000, True),
    (123, False),
    (3600, True),
    (1800, True),
    (123, False),
    (1800, True),
]


@pytest.mark.parametrize("n_images,store_id", n_images_store_id)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb",
    new=MagicMock(),
)
def test_ispyb_handler_stores_sampleid_for_full_collection_not_screening(
    n_images: int,
    store_id: bool,
    params: RotationScan,
):
    cb = RotationISPyBCallback()
    cb.active = True

    doc: RunStart = {
        "time": 0,
        "uid": "abc123",
    }

    params.sample_id = 987678
    params.scan_width_deg = n_images / 10
    if n_images < 200:
        params.ispyb_experiment_type = IspybExperimentType.CHARACTERIZATION
    assert params.num_images == n_images
    doc["subplan_name"] = CONST.PLAN.ROTATION_OUTER  # type: ignore
    doc["mx_bluesky_parameters"] = params.model_dump_json()  # type: ignore

    cb.start(doc)
    assert (cb.last_sample_id == 987678) is store_id



import os
import sys
from subprocess import CalledProcessError, Popen


def run_process_and_print_output(proc_to_run):
    with Popen(proc_to_run, bufsize=1, universal_newlines=True) as p:
        p.communicate()
    if p.returncode != 0:
        raise CalledProcessError(p.returncode, p.args)


def setup_venv(path_to_create_venv_script, deployment_directory):
    # Set up environment and run /dls_dev_env.sh...
    os.chdir(deployment_directory)
    print(f"Setting up environment in {deployment_directory}")

    run_process_and_print_output(path_to_create_venv_script)


if __name__ == "__main__":
    # This should only be entered from the control machine
    path_to_create_venv_script = sys.argv[1]
    deployment_directory = sys.argv[2]
    setup_venv(path_to_create_venv_script, deployment_directory)



import dataclasses
import json
import re
from os import path
from typing import Any, TypeVar

import requests
from deepdiff.diff import DeepDiff
from dodal.utils import get_beamline_name
from jsonschema import ValidationError
from pydantic_extra_types.semantic_version import SemanticVersion

from mx_bluesky.common.parameters.components import (
    PARAMETER_VERSION,
    MxBlueskyParameters,
    TopNByMaxCountSelection,
    WithCentreSelection,
    WithOptionalEnergyChange,
    WithSample,
    WithVisit,
)
from mx_bluesky.common.parameters.constants import (
    GridscanParamConstants,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.utils import convert_angstrom_to_eV
from mx_bluesky.hyperion.parameters.components import WithHyperionUDCFeatures
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadThenCentre

T = TypeVar("T", bound=WithVisit)
AGAMEMNON_URL = "http://agamemnon.diamond.ac.uk/"
MULTIPIN_PREFIX = "multipin"
MULTIPIN_FORMAT_DESC = "Expected multipin format is multipin_{number_of_wells}x{well_size}+{distance_between_tip_and_first_well}"
MULTIPIN_REGEX = rf"^{MULTIPIN_PREFIX}_(\d+)x(\d+(?:\.\d+)?)\+(\d+(?:\.\d+)?)$"
MX_GENERAL_ROOT_REGEX = r"^/dls/(?P<beamline>[^/]+)/data/[^/]*/(?P<visit>[^/]+)(?:/|$)"


class AgamemnonLoadCentreCollect(
    MxBlueskyParameters,
    WithVisit,
    WithSample,
    WithCentreSelection,
    WithHyperionUDCFeatures,
    WithOptionalEnergyChange,
):
    """Experiment parameters to compare against GDA populated LoadCentreCollect."""

    robot_load_then_centre: RobotLoadThenCentre


@dataclasses.dataclass
class PinType:
    expected_number_of_crystals: int
    single_well_width_um: float
    tip_to_first_well_um: float = 0

    @property
    def full_width(self) -> float:
        """This is the "width" of the area where there may be samples.

        From a pin perspective this is along the length of the pin but we use width here as
        we mount the sample at 90 deg to the optical camera.

        We calculate the full width by adding all the gaps between wells then assuming
        there is a buffer of {tip_to_first_well_um} either side too. In reality the
        calculation does not need to be very exact as long as we get a width that's good
        enough to use for optical centring and XRC grid size.
        """
        return (self.expected_number_of_crystals - 1) * self.single_well_width_um + (
            2 * self.tip_to_first_well_um
        )


class SinglePin(PinType):
    def __init__(self):
        super().__init__(1, GridscanParamConstants.WIDTH_UM)

    @property
    def full_width(self) -> float:
        return self.single_well_width_um


def _get_parameters_from_url(url: str) -> dict:
    response = requests.get(url, headers={"Accept": "application/json"})
    response.raise_for_status()
    response_json = json.loads(response.content)
    try:
        return response_json["collect"]
    except KeyError as e:
        raise KeyError(f"Unexpected json from agamemnon: {response_json}") from e


def get_pin_type_from_agamemnon_parameters(parameters: dict) -> PinType:
    loop_type_name: str | None = parameters["sample"]["loopType"]
    if loop_type_name:
        regex_search = re.search(MULTIPIN_REGEX, loop_type_name)
        if regex_search:
            wells, well_size, tip_to_first_well = regex_search.groups()
            return PinType(int(wells), float(well_size), float(tip_to_first_well))
        else:
            loop_type_message = (
                f"Agamemnon loop type of {loop_type_name} not recognised"
            )
            if loop_type_name.startswith(MULTIPIN_PREFIX):
                raise ValueError(f"{loop_type_message}. {MULTIPIN_FORMAT_DESC}")
            LOGGER.warning(f"{loop_type_message}, assuming single pin")
    return SinglePin()


def get_next_instruction(beamline: str) -> dict:
    return _get_parameters_from_url(AGAMEMNON_URL + f"getnextcollect/{beamline}")


def get_withvisit_parameters_from_agamemnon(parameters: dict) -> tuple:
    try:
        prefix = parameters["prefix"]
        collection = parameters["collection"]
        # Assuming distance is identical for multiple collections. Remove after https://github.com/DiamondLightSource/mx-bluesky/issues/773
        detector_distance = collection[0]["distance"]
    except KeyError as e:
        raise KeyError("Unexpected json from agamemnon") from e

    match = re.match(MX_GENERAL_ROOT_REGEX, prefix) if prefix else None

    if match:
        return (match.group("visit"), detector_distance)

    raise ValueError(
        f"Agamemnon prefix '{prefix}' does not match MX-General root structure"
    )


def get_withsample_parameters_from_agamemnon(parameters: dict) -> dict[str, Any]:
    assert parameters.get("sample"), "instruction does not have a sample"
    return {
        "sample_id": parameters["sample"]["id"],
        "sample_puck": parameters["sample"]["container"],
        "sample_pin": parameters["sample"]["position"],
    }


def get_withenergy_parameters_from_agamemnon(parameters: dict) -> dict[str, Any]:
    try:
        first_collection: dict = parameters["collection"][0]
        wavelength = first_collection.get("wavelength")
        assert isinstance(wavelength, float)
        demand_energy_ev = convert_angstrom_to_eV(wavelength)
        return {"demand_energy_ev": demand_energy_ev}
    except (KeyError, IndexError, AttributeError, TypeError):
        return {"demand_energy_ev": None}


def get_param_version() -> SemanticVersion:
    return SemanticVersion.validate_from_str(str(PARAMETER_VERSION))


def create_robot_load_then_centre_params_from_agamemnon(
    parameters: dict,
) -> RobotLoadThenCentre:
    visit, detector_distance = get_withvisit_parameters_from_agamemnon(parameters)
    with_sample_params = get_withsample_parameters_from_agamemnon(parameters)
    with_energy_params = get_withenergy_parameters_from_agamemnon(parameters)
    visit_directory, file_name = path.split(parameters["prefix"])
    return RobotLoadThenCentre(
        parameter_model_version=get_param_version(),
        storage_directory=visit_directory + "/xraycentring",
        visit=visit,
        detector_distance_mm=detector_distance,
        snapshot_directory=visit_directory + "/snapshots",
        file_name=file_name,
        **with_energy_params,
        **with_sample_params,
    )


def populate_parameters_from_agamemnon(agamemnon_params):
    visit, detector_distance = get_withvisit_parameters_from_agamemnon(agamemnon_params)
    with_sample_params = get_withsample_parameters_from_agamemnon(agamemnon_params)
    with_energy_params = get_withenergy_parameters_from_agamemnon(agamemnon_params)
    pin_type = get_pin_type_from_agamemnon_parameters(agamemnon_params)
    robot_load_params = create_robot_load_then_centre_params_from_agamemnon(
        agamemnon_params
    )
    return AgamemnonLoadCentreCollect(
        parameter_model_version=SemanticVersion.validate_from_str(
            str(PARAMETER_VERSION)
        ),
        visit=visit,
        detector_distance_mm=detector_distance,
        select_centres=TopNByMaxCountSelection(n=pin_type.expected_number_of_crystals),
        robot_load_then_centre=robot_load_params,
        **with_sample_params,
        **with_energy_params,
    )


def create_parameters_from_agamemnon() -> AgamemnonLoadCentreCollect:
    beamline_name = get_beamline_name("i03")
    agamemnon_params = get_next_instruction(beamline_name)

    return populate_parameters_from_agamemnon(agamemnon_params)


def compare_params(load_centre_collect_params):
    try:
        parameters = create_parameters_from_agamemnon()

        # Log differences against GDA populated parameters
        differences = DeepDiff(
            parameters, load_centre_collect_params, math_epsilon=1e-5
        )
        if differences:
            LOGGER.info(
                f"Different parameters found when directly reading from Hyperion: {differences}"
            )
    except (ValueError, KeyError) as e:
        LOGGER.warning(f"Failed to compare parameters: {e}")
    except Exception as e:
        LOGGER.warning(f"Unexpected error occurred. Failed to compare parameters: {e}")


def update_params_from_agamemnon(parameters: T) -> T:
    try:
        beamline_name = get_beamline_name("i03")
        agamemnon_params = get_next_instruction(beamline_name)
        pin_type = get_pin_type_from_agamemnon_parameters(agamemnon_params)
        if isinstance(parameters, LoadCentreCollect):
            parameters.robot_load_then_centre.tip_offset_um = pin_type.full_width / 2
            parameters.robot_load_then_centre.grid_width_um = pin_type.full_width
            parameters.select_centres.n = pin_type.expected_number_of_crystals
            if pin_type != SinglePin():
                # Snapshots between each collection take a lot of time.
                # Before we do https://github.com/DiamondLightSource/mx-bluesky/issues/226
                # this will give no snapshots but that's preferable
                parameters.multi_rotation_scan.snapshot_omegas_deg = []
    except (ValueError, ValidationError) as e:
        LOGGER.warning(f"Failed to update parameters: {e}")
    except Exception as e:
        LOGGER.warning(f"Unexpected error occurred. Failed to update parameters: {e}")

    return parameters



from pathlib import Path

import pytest

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.beamlines.i24.serial.parameters import (
    DetectorName,
    FixedTargetParameters,
    get_chip_format,
)

TEST_PATH = Path("tests/test_data/test_daq_configuration")

TEST_LUT = {
    DetectorName.EIGER: TEST_PATH / "lookup/test_det_dist_converter.txt",
}


@pytest.fixture
def dummy_params_with_pp():
    oxford_defaults = get_chip_format(ChipType.Oxford)
    params = {
        "visit": "/tmp/dls/i24/fixed/foo",
        "directory": "bar",
        "filename": "chip",
        "exposure_time_s": 0.01,
        "detector_distance_mm": 100,
        "detector_name": "eiger",
        "transmission": 1.0,
        "num_exposures": 1,
        "chip": oxford_defaults.model_dump(),
        "map_type": 1,
        "pump_repeat": 3,
        "checker_pattern": False,
        "chip_map": [1],
        "laser_dwell_s": 0.02,
        "laser_delay_s": 0.05,
    }
    return FixedTargetParameters(**params)



from inspect import getfullargspec

import pytest

import mx_bluesky.hyperion.experiment_plans as plan_module
from mx_bluesky.common.parameters.components import MxBlueskyParameters
from mx_bluesky.hyperion.experiment_plans import __all__ as exposed_plans
from mx_bluesky.hyperion.experiment_plans.experiment_registry import (
    PLAN_REGISTRY,
    do_nothing,
)


def test_experiment_registry_param_types():
    for plan in PLAN_REGISTRY.keys():
        assert issubclass(PLAN_REGISTRY[plan]["param_type"], MxBlueskyParameters)


def test_exposed_plans_in_reg():
    for plan in exposed_plans:
        assert plan in PLAN_REGISTRY.keys()


@pytest.mark.skip(
    reason="Conflicting MsgGenerator import strangeness https://github.com/DiamondLightSource/mx-bluesky/pull/564"
)
def test_param_types_in_registry_match_plan():
    for plan in exposed_plans:
        plan_function = getattr(plan_module, plan)
        plan_args = getfullargspec(plan_function)
        param_arg_type = plan_args.annotations["parameters"]
        assert PLAN_REGISTRY[plan]["param_type"].__name__ in param_arg_type


def test_do_nothing():
    do_nothing()



from dataclasses import asdict, dataclass
from enum import Enum


class Orientation(Enum):
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"


@dataclass()
class DataCollectionGroupInfo:
    visit_string: str
    experiment_type: str
    sample_id: int | None
    sample_barcode: str | None = None
    comments: str | None = None


@dataclass(kw_only=True)
class DataCollectionInfo:
    omega_start: float | None = None
    data_collection_number: int | None = None
    xtal_snapshot1: str | None = None
    xtal_snapshot2: str | None = None
    xtal_snapshot3: str | None = None
    xtal_snapshot4: str | None = None

    n_images: int | None = None
    axis_range: float | None = None
    axis_end: float | None = None
    kappa_start: float | None = None

    parent_id: int | None = None
    visit_string: str | None = None
    sample_id: int | None = None
    detector_id: int | None = None
    axis_start: float | None = None
    focal_spot_size_at_samplex: float | None = None
    focal_spot_size_at_sampley: float | None = None
    slitgap_vertical: float | None = None
    slitgap_horizontal: float | None = None
    beamsize_at_samplex: float | None = None
    beamsize_at_sampley: float | None = None
    transmission: float | None = None
    comments: str | None = None
    detector_distance: float | None = None
    exp_time: float | None = None
    imgdir: str | None = None
    file_template: str | None = None
    imgprefix: str | None = None
    imgsuffix: str | None = None
    n_passes: int | None = None
    overlap: int | None = None
    flux: float | None = None
    start_image_number: int | None = None
    resolution: float | None = None
    wavelength: float | None = None
    xbeam: float | None = None
    ybeam: float | None = None
    synchrotron_mode: str | None = None
    undulator_gap1: float | None = None
    start_time: str | None = None


@dataclass
class DataCollectionPositionInfo:
    pos_x: float
    pos_y: float
    pos_z: float


@dataclass
class DataCollectionGridInfo:
    """This information is used by Zocalo gridscan per-image-analysis"""

    dx_in_mm: float
    dy_in_mm: float
    steps_x: int
    steps_y: int
    microns_per_pixel_x: float
    microns_per_pixel_y: float
    snapshot_offset_x_pixel: int
    snapshot_offset_y_pixel: int
    orientation: Orientation
    snaked: bool

    def as_dict(self):
        d = asdict(self)
        d["orientation"] = self.orientation.value
        return d


@dataclass(kw_only=True)
class ScanDataInfo:
    data_collection_info: DataCollectionInfo
    data_collection_id: int | None = None
    data_collection_position_info: DataCollectionPositionInfo | None = None
    data_collection_grid_info: DataCollectionGridInfo | None = None



import pytest
from tests.unit_tests.hyperion.experiment_plans.conftest import (
    FLYSCAN_RESULT_HIGH,
    FLYSCAN_RESULT_LOW,
    FLYSCAN_RESULT_MED,
)

from mx_bluesky.common.xrc_result import (
    top_n_by_max_count,
)


@pytest.mark.parametrize(
    "input_sequence, expected_sequence, n",
    [
        [
            [FLYSCAN_RESULT_LOW, FLYSCAN_RESULT_MED, FLYSCAN_RESULT_HIGH],
            [FLYSCAN_RESULT_HIGH, FLYSCAN_RESULT_MED],
            2,
        ],
        [[FLYSCAN_RESULT_LOW], [FLYSCAN_RESULT_LOW], 2],
    ],
)
def test_top_n_by_max_count(input_sequence, expected_sequence, n):
    actual_sequence = top_n_by_max_count(input_sequence, n)
    assert actual_sequence == expected_sequence



from unittest.mock import MagicMock

import pytest
from bluesky import plan_stubs as bps
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from ophyd.status import Status
from ophyd_async.sim import SimMotor

from mx_bluesky.hyperion.device_setup_plans.setup_oav import (
    pre_centring_setup_oav,
)

OAV_CENTRING_JSON = "tests/test_data/test_OAVCentring.json"


@pytest.fixture
def mock_parameters():
    return OAVParameters("loopCentring", OAV_CENTRING_JSON)


def test_when_set_up_oav_then_only_waits_on_oav_to_finish(
    mock_parameters: OAVParameters, oav: OAV, ophyd_pin_tip_detection: PinTipDetection
):
    """This test will hang if pre_centring_setup_oav waits too generally as my_waiting_device
    never finishes moving"""
    my_waiting_device = SimMotor(name="")
    my_waiting_device.set = MagicMock(return_value=Status())

    def my_plan():
        yield from bps.abs_set(my_waiting_device, 10, wait=False)
        yield from pre_centring_setup_oav(oav, mock_parameters, ophyd_pin_tip_detection)

    RE = RunEngine()
    RE(my_plan())



from pathlib import Path

# import pytest
from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    ChipDescription,
    ExtruderParameters,
    FixedTargetParameters,
    SSXType,
)
from mx_bluesky.beamlines.i24.serial.parameters.utils import get_chip_format


def test_extruder_params(dummy_params_ex):
    assert isinstance(dummy_params_ex, ExtruderParameters)
    assert dummy_params_ex.collection_directory == Path("/tmp/dls/i24/extruder/foo/bar")
    assert dummy_params_ex.nexgen_experiment_type == "extruder"
    assert dummy_params_ex.ispyb_experiment_type is SSXType.EXTRUDER


def test_chip_params_with_mapping_lite_and_one_block(dummy_params_without_pp):
    assert isinstance(dummy_params_without_pp, FixedTargetParameters)
    assert dummy_params_without_pp.total_num_images == 400
    assert dummy_params_without_pp.chip.chip_type is ChipType.Oxford
    assert dummy_params_without_pp.nexgen_experiment_type == "fixed-target"
    assert dummy_params_without_pp.ispyb_experiment_type is SSXType.FIXED


def test_chip_params_with_multiple_blocks(dummy_params_without_pp):
    dummy_params_without_pp.chip_map = [1, 2, 3]
    assert dummy_params_without_pp.total_num_images == 1200
    assert dummy_params_without_pp.map_type is MappingType.Lite
    assert dummy_params_without_pp.pump_repeat is PumpProbeSetting.NoPP


def test_chip_params_with_multiple_exposures(dummy_params_without_pp):
    dummy_params_without_pp.num_exposures = 2
    dummy_params_without_pp.chip_map = [1, 16]
    assert dummy_params_without_pp.total_num_images == 1600


def test_chip_params_with_no_mapping_for_oxford_chip():
    oxford_defaults = get_chip_format(ChipType.Oxford)
    oxford_chip = {
        "visit": "foo",
        "directory": "bar",
        "filename": "chip",
        "exposure_time_s": 0.01,
        "detector_distance_mm": 100,
        "detector_name": "eiger",
        "transmission": 1.0,
        "num_exposures": 1,
        "chip": oxford_defaults.model_dump(),
        "map_type": 0,
        "pump_repeat": 0,
        "checker_pattern": False,
        "chip_map": [],
    }
    params = FixedTargetParameters(**oxford_chip)
    assert params.total_num_images == 25600


def test_chip_params_with_no_mapping_for_custom_chip():
    custom_defaults = {
        "chip_type": ChipType.Custom,
        "x_num_steps": 5,
        "y_num_steps": 3,
        "x_step_size": 0.1,
        "y_step_size": 0.1,
        "x_blocks": 1,
        "y_blocks": 1,
        "b2b_horz": 0.0,
        "b2b_vert": 0.0,
    }
    custom_chip = {
        "visit": "foo",
        "directory": "bar",
        "filename": "chip",
        "exposure_time_s": 0.01,
        "detector_distance_mm": 100,
        "detector_name": "eiger",
        "transmission": 1.0,
        "num_exposures": 2,
        "chip": ChipDescription(**custom_defaults),
        "map_type": 0,
        "pump_repeat": 0,
        "checker_pattern": False,
        "chip_map": [],
    }
    params = FixedTargetParameters(**custom_chip)
    assert params.total_num_images == 30



from bluesky.callbacks import CallbackBase
from event_model import RunStart, RunStop

from mx_bluesky.common.utils.log import set_uid_tag


class LogUidTaggingCallback(CallbackBase):
    def __init__(self) -> None:
        """Sets the logging filter to add the outermost run uid to graylog messages"""
        self.run_uid = None

    def start(self, doc: RunStart):
        if self.run_uid is None:
            self.run_uid = doc.get("uid")
            set_uid_tag(self.run_uid)

    def stop(self, doc: RunStop):
        if doc.get("run_start") == self.run_uid:
            self.run_uid = None
            set_uid_tag(None)



import re

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.zocalo import ZOCALO_READING_PLAN_NAME, ZocaloResults
from dodal.utils import is_test_mode

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import (
    EnvironmentConstants,
    PlanNameConstants,
)
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_gridscan_callbacks,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan
from tests.conftest import create_dummy_scan_spec

"""
If fake-zocalo system tests are failing, check that the RMQ instance is set up right:

- Open the RMQ webpage specified when you start the fake zocalo and login with the
provided credentials

- go to the admin panel and under the 'exchanges' tab ensure that there is a 'results'
exchange for the zocalo vhost (other settings can be left blank)

- go to the 'queues and streams' tab and add a binding for the xrc.i03 queue to the
results exchange, with the routing key 'xrc.i03'

- make sure that there are no un-acked/un-delivered messages on the i03.xrc queue
"""


@bpp.set_run_key_decorator("testing125")
@bpp.run_decorator(
    md={
        "subplan_name": PlanNameConstants.DO_FGS,
        "zocalo_environment": EnvironmentConstants.ZOCALO_ENV,
        "scan_points": create_dummy_scan_spec(10, 20, 30),
    }
)
def fake_fgs_plan():
    yield from bps.sleep(0)


@pytest.fixture
def run_zocalo_with_dev_ispyb(
    dummy_params: HyperionSpecifiedThreeDGridScan,
    dummy_ispyb_3d,
    RE: RunEngine,
    zocalo_for_fake_zocalo: ZocaloResults,
):
    async def inner(sample_name="", fallback=np.array([0, 0, 0])):
        dummy_params.file_name = sample_name
        _, ispyb_callback = create_gridscan_callbacks()
        RE.subscribe(ispyb_callback)

        @bpp.set_run_key_decorator("testing123")
        def trigger_zocalo_after_fast_grid_scan():
            @bpp.set_run_key_decorator("testing124")
            @bpp.stage_decorator([zocalo_for_fake_zocalo])
            @bpp.run_decorator(
                md={
                    "subplan_name": CONST.PLAN.GRIDSCAN_OUTER,
                    CONST.TRIGGER.ZOCALO: PlanNameConstants.DO_FGS,
                    "zocalo_environment": EnvironmentConstants.ZOCALO_ENV,
                    "mx_bluesky_parameters": dummy_params.model_dump_json(),
                }
            )
            def inner_plan():
                yield from fake_fgs_plan()
                yield from bps.trigger_and_read(
                    [zocalo_for_fake_zocalo], name=ZOCALO_READING_PLAN_NAME
                )

            yield from inner_plan()

        RE(
            ispyb_activation_wrapper(
                trigger_zocalo_after_fast_grid_scan(), dummy_params
            )
        )
        centre = await zocalo_for_fake_zocalo.centre_of_mass.get_value()
        if centre.size == 0:
            centre = fallback
        else:
            centre = centre[0]

        return ispyb_callback, ispyb_callback.emit_cb, centre

    return inner


@pytest.mark.system_test
def test_is_test_mode():
    assert is_test_mode(), "DODAL_TEST_MODE must be set in environment before launch"


@pytest.mark.system_test
async def test_given_a_result_with_no_diffraction_when_zocalo_called_then_move_to_fallback(
    run_zocalo_with_dev_ispyb,
):
    fallback = np.array([1, 2, 3])
    _, _, centre = await run_zocalo_with_dev_ispyb("NO_DIFF", fallback)
    assert np.allclose(centre, fallback)


@pytest.mark.system_test
async def test_given_a_result_with_no_diffraction_ispyb_comment_updated(
    run_zocalo_with_dev_ispyb, fetch_comment
):
    ispyb, zc, _ = await run_zocalo_with_dev_ispyb("NO_DIFF")

    comment = fetch_comment(ispyb.ispyb_ids.data_collection_ids[0])
    assert "Zocalo found no crystals in this gridscan." in comment


@pytest.mark.system_test
async def test_zocalo_adds_nonzero_comment_time(
    run_zocalo_with_dev_ispyb, fetch_comment
):
    ispyb, zc, _ = await run_zocalo_with_dev_ispyb()

    comment = fetch_comment(ispyb.ispyb_ids.data_collection_ids[0])
    match = re.match(r".*Zocalo processing took (\d+\.\d+) s", comment)
    assert match
    time_s = float(match.group(1))
    assert time_s > 0
    assert time_s < 180


@pytest.mark.system_test
async def test_given_a_single_crystal_result_ispyb_comment_updated(
    run_zocalo_with_dev_ispyb, fetch_comment
):
    ispyb, zc, _ = await run_zocalo_with_dev_ispyb()
    comment = fetch_comment(ispyb.ispyb_ids.data_collection_ids[0])
    assert "Crystal 1" in comment
    assert "Strength" in comment
    assert "Size (grid boxes)" in comment


@pytest.mark.system_test
async def test_given_a_result_with_multiple_crystals_ispyb_comment_updated(
    run_zocalo_with_dev_ispyb, fetch_comment
):
    ispyb, zc, _ = await run_zocalo_with_dev_ispyb("MULTI_X")

    comment = fetch_comment(ispyb.ispyb_ids.data_collection_ids[0])
    assert "Crystal 1" and "Crystal 2" in comment
    assert "Strength" in comment
    assert "Position (grid boxes)" in comment
